/*
 *  ir02
 *  参考：StreamHTTPClient.ino, SPI_Multiple_Buses.ino, SD_test.ino, SimpleTime.ino, GPIOInterrupt
 *    Arduino IDE 設定
 *      ボード：ESP32 Dev Module
 *      Upload speed : 921600
 *      CPU Freq : 240MHz
 *      Flash Freq : 80MHz
 *      Flash Mode : QIO
 *      Flash Size : 4MB
 *      Partition Scheme : Default 4MB with spiffs
 *      Psram : disabled
 */
#include <Arduino.h>

#include <WiFi.h>
#include <WiFiMulti.h>
#include <HTTPClient.h>
#include <SPI.h>
#include <SD.h>
#include <FS.h>
#include <Wire.h>           // I2C
#include <Ticker.h>

//#include "vs1053b-patches.h"

#define RECVBUFF_SIZE       (7*1024*10)
#define RECVUNIT            (128*1)
#define M3UBUFF_SIZE        (1024)
#define SKIPBUFF_SIZE       512
#define FILE_SIZE           128
#define URL_SIZE            256
#define BUF_SIZE            512
#define M3U_TABLE_SIZE      12
#define M3U_NHK             1
#define M3U_MUSICBIRD       2
#define M3U_GET_OK          0
#define M3U_GET_ERROR       1
#define M3U_GET_TOO_LARGE   2
#define STATIONID_SIZE      3
#define STATIONNAME_SIZE    16
#define STATIONLIST_SIZE    20
#define FILEPATH_SIZE        16      // ルート直下のファイルのみ想定
#define RECVMODE_LISTEN     1       // リアルタイム受信
#define RECVMODE_SCHREC     2       // 予約録音
#define GETSOUND_TIMEOUT        8000    // ms
#define MAX_RETRY_COUNT         10
#define MAX_RETRY_COUNT_M3U     5
#define RETRY_WAIT_TIME         10000   // ms
#define ID_REPLAY               "ZZZ"   // Station ID for REPLAY
#define RECFILELIST_SIZE        30
#define CHATTERING_ADJUST       15
#define VS1053_DECODE_BYTE      32
#define TICKER_INTERVAL         100
#define BLINK_CYCLE_SETUP       800
#define SCHREC_POWERUP_MARGIN   2       // 予約録音時、起動時刻マージン(分)
#define AUTO_POFF_TIME          (5*60*1000)     // ms
#define MP3_HDR_BYTE            4       // MP3 フレームヘッダバイト数
#define AAC_HDR_BYTE            7

#define WIFI_SETTING_FILE       "/WIFI02.TXT"           // WIFI設定
#define STATIONLIST_FILE        "/STATIN02.TXT"         // 局リスト
#define SCHEDULE_FILE           "/SCHDUL02.TXT"         // タイマー録音スケジュール
#define NEXT_SCHEDULE_FILE      "/NXTSCH02.TXT"         // 次回録音（本プログラムが作成）
#define VOLUME_FILE             "/VOLUME02.DAT"         // ボリューム値保存ファイル
#define LOG_FILE                "/LOG02.TXT"            // ログファイル

// Station ListのMETHOD
#define METHOD_M3U              "M3U"
#define METHOD_NHK              "NHK"
#define METHOD_MUSICBIRD        "MUSICBIRD"
#define METHOD_HIGH             "HIGH"
#define METHOD_NORMAL           "NORMAL"

// VS1053 Interface pins
#define MP3_MISO        21
#define MP3_MOSI        23
#define MP3_CLK         19
#define MP3_RESET       17
#define MP3_CS          22
#define MP3_DCS         18
#define MP3_DREQ        5

// VS1053 register constants
#define SCI_MODE                0x00
#define SCI_STATUS              0x01
#define SCI_BASS                0x02
#define SCI_CLOCKF              0x03
#define SCI_AUDATA              0x05
#define SCI_WRAM                0x06
#define SCI_WRAMADDR            0x07
#define SCI_HDAT0               0x08
#define SCI_HDAT1               0x09
#define SCI_AIADDR              0x0A
#define SCI_VOL                 0x0B
#define SCI_AICTRL0             0x0C
#define SCI_AICTRL1             0x0D
#define SCI_AICTRL2             0x0E
#define SCI_AICTRL3             0x0F

// VS1053 SCI_MODE bits
#define SM_DIFF_B               0x0001    // Bit:0
#define SM_LAYER12_B            0x0002    /* VS1063, VS1053, VS1033, VS1011 */
#define SM_RESET_B              0x0004    // Soft reset
#define SM_CANCEL_B             0x0008    // Bit:3
#define SM_TESTS_B              0x0020    // Bit:5
#define SM_STREAM_B             0x0040    // Bit:6
#define SM_DACT_B               0x0100      
#define SM_SDIORD_B             0x0200    // Bit:9
#define SM_SDISHARE_B           0x0400    // Bit:10
#define SM_SDINEW_B             0x0800    // Bit:11
#define SM_ENCODE_B             0x1000    // Bit:12 /* VS1063 */
#define SM_LINE1_B              0x4000    // Bit:14 /* VS1063, VS1053 */
#define SM_CLK_RANGE_B          0x8000    // Bit:15 /* VS1063, VS1053, VS1033 */

// VS1053 settings
#define CLOCKF_FREQ             ((12000000 - 8000000)/4000)   // (1000)  (Crystal=12MHz)
#define SC_ADD                  0x1800

// SDピン定義
#define SD_CLK          14
#define SD_MOSI         15
#define SD_MISO         4
#define SD_SS           13

// I2Cピン定義
#define I2C_SCL         32
#define IC2_SDA         33

// その他ピン定義
#define POWER_LED       16
#define POWER_ON        12
#define POWER_OFF       34
#define SW_SEL          39
#define SW_RETURN       35
#define SW_UP           25
#define SW_DOWN         26
#define SW_OPTION       36
#define AMP_CE          27

// イベント
#define SWSELECT_PRESSED        0x82
#define SWRETURN_PRESSED        0x83
#define SWUP_PRESSED            0x84
#define SWDOWN_PRESSED          0x85
#define SWPOFF_PRESSED          0x86
#define SWOPTION_PRESSED        0x87

#define LABEL_IR02              "IR02    "
#define LABEL_RECIEVE           "Rcv     "
#define LABEL_RECIEVE_RECORDING "Rcv Rec "
#define LABEL_REPLAY            "Replay  "
#define LABEL_CLEAR             "        "

// I2C/RTC
#define RTC_ADDRESS     0x68
const long  gmtOffset_sec = (9*3600);
const int   daylightOffset_sec = 0;
// I2C/LCD
#define LCD_ADDRESS      0x3e
#define LCD_CONTRAST    35          // コントラスト(0～63)

// WIFI
WiFiMulti   wifiMulti;
char WifiSsid[32];
char WifiPassword[32];
char NtpServer[32];

uint8_t     RecvBuff[RECVBUFF_SIZE] = {0};
int  dat_start;    // 再生データ開始＝空き最後
int  emp_start;    // 再生データ最後＝空き開始
volatile int  dat_size;     // 再生可能データバイト数
int  RecvMode;           // 受信モード
int  SelectedMenu = 0;
int    SelectedFile;
volatile bool Decoding = false;        // isr_dreq()からDecode()を実行中
bool EnableDreqInt = false;              // 音声データなどの受信中
//bool DreqEvent = false;
bool BlinkLed = false;
int BlinkCycle;
int BlinkCycle2;
unsigned int  BlinkCnt = 0;
unsigned long    RecFileBytes;           // 録音先ファイル　バイト数
//int SampleRate;
//int BitRate;
int bitrate_tab[] = {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 999};
int sample_tab[] = {44100, 48000, 32000, 99999};
int AdtsSampling[] = {96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000, 12000, 11025, 8000, 7350 };
int NumAdtsSampling = (sizeof AdtsSampling / sizeof AdtsSampling[0]);
//double SecFrame;
//uint8_t    FrameByte[2];
bool SaveAtDecode;

// VS1053/SPI
#define MIN_VOLUME              0x00
#define MAX_VOLUME              0xD0
#define DEFAULT_VOLUME          0x30
#define STEP_VOLUME             0x06
uint16_t     Volume;
uint16_t     VolumeBak;
const int spiClk = 600000; // 600khz
SPIClass * Spi = NULL;
// SD/SPI
SPIClass * SpiSd = NULL;
File    FileSnd;
File    FileObj;
File    FileLog;
// RTC
uint8_t     RtcBuff[16];     // mp3_05/m41t62.c

// 局リスト構造体
typedef struct StationList {
    char    id[STATIONID_SIZE + 1];
    char    name[STATIONNAME_SIZE + 1];
} STATIONLIST;

STATIONLIST StList[STATIONLIST_SIZE];    // 局リスト
int NumStList;   // 局数

// 録音スケジュール
typedef struct schedule {
    char        sta[4];     // Station ID
    int         year;
    int         month;
    int         day;
    int         hour;
    int         min;
    int         len;
} SCHEDULE;

// 録音済みファイル
typedef struct replay_file {
    char        file[FILEPATH_SIZE];
    size_t      size;       // バイト
    int         min;        // 分
} REPLAY_FILE;

REPLAY_FILE RecFileList[RECFILELIST_SIZE];    // 録音済みファイルリスト
int NumRecFileList;   // 同　ファイル数

// M3Uファイル構造体
typedef struct M3Utable {
    char    reqfile[URL_SIZE];     // ファイル名またはURL
    int     status;     // 0:empty, 1:before get 2:after get
} M3UTABLE;

M3UTABLE M3uTable[M3U_TABLE_SIZE];
int NumM3uTable = 0;            // M3uTable内の有効項目数
int M3uTableInsIndex = 0;       // M3uTable内の新規項目挿入位置
int M3uTableGetIndex = 0;       // M3uTable内の次回音源データ取得対象位置

char FileBuf[BUF_SIZE];

Ticker ticker;
volatile bool EVENT_SWPOFF;
volatile bool EVENT_SWSELECT;
volatile bool EVENT_SWRETURN;
volatile bool EVENT_SWUP;
volatile bool EVENT_SWDOWN;
volatile bool EVENT_SWOPTION;
int     CHATTERING_TMR;


// DREQの立ち上がりイベント
void IRAM_ATTR isr_dreq() {
    if (EnableDreqInt == true){
        Decoding = true;
        Decode();
        Decoding = false;
        //DreqEvent = true;
    }
}

// TICKERイベント
void IRAM_ATTR isr_ticker() {
    if (CHATTERING_TMR > 0){
        CHATTERING_TMR++;
        if (CHATTERING_TMR == CHATTERING_ADJUST) CHATTERING_TMR = 0;
    } else {
        if (digitalRead(POWER_OFF) == 0 && EVENT_SWPOFF == false){
            EVENT_SWPOFF = true;
            CHATTERING_TMR = 1;     // chattering timer start
        } else if (digitalRead(SW_SEL) == 0 && EVENT_SWSELECT == false){
            EVENT_SWSELECT = true;
            CHATTERING_TMR = 1;     // chattering timer start
        } else if (digitalRead(SW_RETURN) == 0 && EVENT_SWRETURN == false){
            EVENT_SWRETURN = true;
            CHATTERING_TMR = 1;     // chattering timer start
        } else if (digitalRead(SW_DOWN) == 0 && EVENT_SWUP == false){
            EVENT_SWUP = true;
            CHATTERING_TMR = 1;     // chattering timer start
        } else if (digitalRead(SW_UP) == 0 && EVENT_SWDOWN == false){
            EVENT_SWDOWN = true;
            CHATTERING_TMR = 1;     // chattering timer start
        } else if (digitalRead(SW_OPTION) == 0 && EVENT_SWOPTION == false){
            EVENT_SWOPTION = true;
            CHATTERING_TMR = 1;     // chattering timer start
        }
    }
    
    if (BlinkLed == true){
        int mod = (BlinkCnt % BlinkCycle2 /*14*/);
        if (mod < BlinkCycle /*7*/){
            digitalWrite(POWER_LED, LOW);
        } else {
            digitalWrite(POWER_LED, HIGH);
        }
        BlinkCnt++;
    }
}

void setup() {
    int     retry_cnt = 0;
    wl_status_t wi_status;
    
    pinMode(POWER_ON, OUTPUT);
    digitalWrite(POWER_ON, HIGH);
    pinMode(POWER_LED, OUTPUT);
    digitalWrite(POWER_LED, HIGH);   // turn the LED on (HIGH is the voltage level)
    pinMode(POWER_OFF, INPUT);
    pinMode(SW_SEL, INPUT);
    pinMode(SW_RETURN, INPUT);
    pinMode(SW_DOWN, INPUT);
    pinMode(SW_UP, INPUT);
    pinMode(SW_OPTION, INPUT);
    pinMode(AMP_CE, OUTPUT);
    digitalWrite(AMP_CE, LOW);
    
    ticker.attach_ms(TICKER_INTERVAL, isr_ticker);
    led_blink_start(BLINK_CYCLE_SETUP);
    
    // put your setup code here, to run once:
    Serial.begin(115200);
    delay(1000);
    
    // I2C/LCD
    Wire.setPins(IC2_SDA, I2C_SCL);
    Wire.begin();
    LCD_Setup();
    LCD_Println(0, LABEL_IR02);
   
    // SPI/SD
    SpiSd = new SPIClass(HSPI);
    SpiSd->begin(SD_CLK, SD_MISO, SD_MOSI); //SCLK, MISO, MOSI, SS
    if (!SD.begin(SD_SS, *SpiSd)){
        LCD_Println(0, "SdFailed");
        Serial.print("Card Mount Failed\n");
        PowerDown();
    } else {
        uint8_t cardType = SD.cardType();
        if(cardType == CARD_NONE){
            LCD_Println(0, "No SD   ");
            Serial.print("No SD card attached\n");
            PowerDown();
        }
        Serial.print("Card Mount success\n");
    }
    ReadWifiSetting();
    if (ReadStationList() == 0){
        LCD_Println(0, "Station?");
        delay(5000);
    }
    LCD_Println(0, LABEL_IR02);
    FileLog = SD.open(LOG_FILE, FILE_APPEND);
    
    // RTC
    RTC_Init();
    //RTC_Reset();
    if (RTC_GetAF() == 1){      // Alarm起動したとき
        RecvMode = RECVMODE_SCHREC;
        LCD_Println(1, "AF ON   ");
        Serial.println("AF ON");
    } else {
        RecvMode = RECVMODE_LISTEN;
        LCD_Println(1, "AF OFF  ");
        Serial.println("AF OFF");
    }
    if (RTC_GetAF() == 1){      // AFは0になっているはず。念のためチェック
        LCD_Println(1, "AF ON !?");
        Serial.println("af on !");
        delay(5000);
    }
    DispCurrentDateTime();
    
    wifiMulti.addAP(WifiSsid, WifiPassword);
    while (wifiMulti.run() != WL_CONNECTED) {
        Serial.printf("WAIT FOR CONNECT ...\n");
        LOG_1("WIFI connect retry");
        delay(5000);        // この値が小さいと接続に失敗する(?)
        retry_cnt++;
        if (retry_cnt == MAX_RETRY_COUNT){
            LOG_1("WIFI connect TIMEOUT");
            PowerDown();
        } else {
        }
    }
        
    LCD_Println(1, "WIFI OK ");
    
    // SPI/VS1053
    Spi = new SPIClass(VSPI);
    Spi->begin(MP3_CLK, MP3_MISO, MP3_MOSI); //SCLK, MISO, MOSI, SS
    pinMode(MP3_CS, OUTPUT);
    pinMode(MP3_RESET, OUTPUT);
    pinMode(MP3_DCS, OUTPUT);
    pinMode(MP3_DREQ, INPUT);
    
    
    // VS1053初期化
    LoadVolume();
    VS1053_HardwareReset();
    VS1053_SetVolume(Volume);
    LCD_Println(1, "VS1053  ");
//    VS1053_SineTest();
    
    // 割り込み初期化
    Serial.print("attachInterrupt begin\n");
    attachInterrupt(MP3_DREQ, isr_dreq, RISING);
    Serial.print("attachInterrupt end\n");
    
    EVENT_SWPOFF = false;
    EVENT_SWSELECT = false;
    EVENT_SWRETURN = false;
    EVENT_SWUP = false;
    EVENT_SWDOWN = false;
    EVENT_SWOPTION = false;
    CHATTERING_TMR = 0;
    
    led_blink_stop();
}

/*
 *  局リスト取得
 */
int ReadStationList(){
    char    *c1, *c2;
    int     len;
    int     i;
    
    FileObj = SD.open(STATIONLIST_FILE);
    if(!FileObj){
        LOG_1("Failed to open STATION LIST FILE");
        Serial.println("Failed to open file for reading");
        return 0;
    }
    while (ReadLine(FileObj, FileBuf, URL_SIZE) != 0){
        if (FileBuf[0] == '#') continue;
        c1 = strchr(FileBuf, ',');      // comma
        if (c1 != NULL){
            *c1 = '\0';
            strcpy(StList[NumStList].id, FileBuf);          // "ID"
            c1++;
            c2 = strchr(c1, ',');   // comma
            if (c2 != NULL){
                *c2 = '\0';
                strncpy(StList[NumStList].name, c1, STATIONNAME_SIZE);     // Station "NAME"
                StList[NumStList].name[STATIONNAME_SIZE] = '\0';
                len = strlen(StList[NumStList].name);
                for (i=len; i<STATIONNAME_SIZE; i++){
                    StList[NumStList].name[i] = ' ';
                }
                Serial.printf("id=%s name=%s\n", StList[NumStList].id, StList[NumStList].name);
                c1 = c2 + 1;
                c2 = strchr(c1, ',');   // comma
                if (c2 != NULL){
                }
                NumStList++;
                if (NumStList == (STATIONLIST_SIZE - 1)) break;
            }
        }
    }
    FileObj.close();
    
    //
    strcpy(StList[NumStList].id, ID_REPLAY);
    strcpy(StList[NumStList].name, "REPLAY          ");
    NumStList++;
    return NumStList;
}

/*
 *  ファイルから１行取得。行は0x0aで終わることを前提。１バイト文字だけで構成されることを前提。
 */
int ReadLine(File &file, char *buf, int bufsize){
    int     c;
    int     n = 0;      // 文字数
    
    while ((c = file.read()) != -1){
        if (c == 0x0a){
            break;
        } else if (c == 0x0d){
        } else {
            buf[n++] = c;
            buf[n] = '\0';
            if (n == (bufsize - 1)) break;
        }
    }
    
    return n;
}

/*
 *  WIFI設定などをSDカードのファイルから取得
 */
int ReadWifiSetting(){
    int     len;
    char    *c0, *c1, *c2, *c3;
    int     ret = 0;
    
    File file = SD.open(WIFI_SETTING_FILE);
    if(!file){
        Serial.println("Failed to open file for reading");
        return ret;
    }
    
    len = file.size();
    file.read((uint8_t *)FileBuf, len);
    
    c0 = FileBuf;
    if ((c1 = strchr(c0, 0x0d)) == NULL) {
        return ret;
    }
    *c1 = '\0';
    c1 += 2;
    
    if ((c2 = strchr(c1, 0x0d)) == NULL) {
        return ret;
    }
    *c2 = '\0';
    c2 += 2;
    
    if ((c3 = strchr(c2, 0x0d)) == NULL) {
        return ret;
    }
    *c3 = '\0';
    
    strcpy(WifiSsid, c0);
    strcpy(WifiPassword, c1);
    strcpy(NtpServer, c2);
    
    file.close();
    return ret;
}

/*
 *  外部ファイルの録音スケジュールの内、直近のものを取得して設定
 */
void LoadSchedule(){
    uint8_t     dt[6];
    int     len;
    int     i;
    SCHEDULE    w;
    SCHEDULE    wa;
    
    RTC_GetCurrentDateTime(dt);
    memset(&wa, 0, sizeof(SCHEDULE));
    memset(&w, 0, sizeof(SCHEDULE));

    FileObj = SD.open(SCHEDULE_FILE);
    if(!FileObj){
        RTC_ClearAlarm();
        Serial.println("No Timer Schedule");
        return;
    }
    while (ReadLine(FileObj, FileBuf, URL_SIZE) != 0){
        if (FileBuf[0] == '#') continue;
        len = strlen(FileBuf);
        for (i=0; i<len; i++){
            if (FileBuf[i] == ',') FileBuf[i] = ' ';
        }
        Serial.printf("wa schedule %s %d/%d/%d %d:%d %d\n", wa.sta, wa.year, wa.month, wa.day, wa.hour, wa.min, wa.len);
        sscanf(FileBuf, "%s %d %d %d %d %d %d", w.sta, &w.year, &w.month, &w.day, &w.hour, &w.min, &w.len);
        Serial.printf("schedule %s %d/%d/%d %d:%d %d\n", w.sta, w.year, w.month, w.day, w.hour, w.min, w.len);
        if (w.sta[0] != '\0' 
            && w.year >= 21
            && w.month >= 1 && w.month <= 12
            && w.day >= 1 && w.day <= 31 
            && w.hour >= 0 && w.hour <= 23
            && w.min >= 0 && w.min <= 59
            && w.len > 0){  // 有効な値であること
            
        Serial.printf("  sta[0]=%d wa.sta[0]=%d \n", w.sta[0], wa.sta[0]);
        Serial.printf("  year =%d %d\n", w.year, dt[0]);
        Serial.printf("  month=%d %d\n", w.month, dt[1]);
        Serial.printf("  day  =%d %d\n", w.day, dt[2]);
        Serial.printf("  hour =%d %d\n", w.hour, dt[3]);
        Serial.printf("  min  =%d %d\n", w.min, dt[4]);
        
            if (w.year > dt[0] ||
                (w.year == dt[0] && w.month > dt[1]) ||
                (w.year == dt[0] && w.month == dt[1] && w.day > dt[2]) ||
                (w.year == dt[0] && w.month == dt[1] && w.day == dt[2] && w.hour > dt[3]) ||
                (w.year == dt[0] && w.month == dt[1] && w.day == dt[2] && w.hour == dt[3] && w.min > dt[4])){   // 現在日時より後
                
                Serial.printf("  wa.sta[0]=%d \n", wa.sta[0]);
                if (wa.sta[0] == 0){
                    memcpy(&wa, &w, sizeof(SCHEDULE));
                    Serial.printf("  2 wa.sta[0]=%d \n", wa.sta[0]);
                } else if (compare_schedule(&w, &wa) == 1){
                    memcpy(&wa, &w, sizeof(SCHEDULE));
                }
            }
        }
    }
    FileObj.close();
    
    if (wa.sta[0] != '\0'){
        FileObj = SD.open(NEXT_SCHEDULE_FILE, FILE_WRITE);
        sprintf(FileBuf, "%s %02d %02d %02d %02d %02d %d\n", wa.sta, wa.year, wa.month, wa.day, wa.hour, wa.min, wa.len);
        len = strlen(FileBuf);
        FileObj.write((uint8_t *)FileBuf, len);
        FileObj.close();
        
        Serial.printf("alarm %s %d/%d/%d %d:%d %d\n", wa.sta, wa.year, wa.month, wa.day, wa.hour, wa.min, wa.len);
        set_startup_time(&wa);
        Serial.printf("alarm %s %d/%d/%d %d:%d %d\n", wa.sta, wa.year, wa.month, wa.day, wa.hour, wa.min, wa.len);
        RTC_SetAlarm(wa.month, wa.day, wa.hour, wa.min, 0);
    } else {
        SD.remove(NEXT_SCHEDULE_FILE);
        RTC_ClearAlarm();
    }
}

/*
 *  予約録音時の起動時刻を取得
 */
void set_startup_time(SCHEDULE *wa){
    if (wa->min >= SCHREC_POWERUP_MARGIN){
        wa->min -= SCHREC_POWERUP_MARGIN;
    } else {    // 0分 ～ SCHREC_POWERUP_MARGIN分
        wa->min = 60 - (SCHREC_POWERUP_MARGIN - wa->min);
        
        if (wa->hour > 0){
            wa->hour -= 1;
        } else {
            wa->hour = 23;
            // 前日
            if (wa->day > 1){
                wa->day -= 1;
            } else {    // 一日
                // 日を前月の最終日
                switch (wa->month){
                case 1: wa->day = 31; break;
                case 2: wa->day = 31; break;
                case 3: wa->day = 28; break;    // うるう年を考慮しない
                case 4: wa->day = 31; break;
                case 5: wa->day = 30; break;
                case 6: wa->day = 31; break;
                case 7: wa->day = 30; break;
                case 8: wa->day = 31; break;
                case 9: wa->day = 31; break;
                case 10: wa->day = 30; break;
                case 11: wa->day = 31; break;
                case 12: wa->day = 30; break;
                }
                
                // 月
                if (wa->month > 1){
                    wa->month -= 1;
                } else {
                    wa->month = 12;
                }
            }
        }
    }
}


/*
 *  スケジュールを比較
 *  s1がs2の前のとき(小さいとき)1を返す
 */
int compare_schedule(SCHEDULE *s1, SCHEDULE *s2){
        Serial.printf("year =%d %d\n", s1->year, s2->year);
        Serial.printf("month=%d %d\n", s1->month, s2->month);
        Serial.printf("day  =%d %d\n", s1->day, s2->day);
        Serial.printf("hour =%d %d\n", s1->hour, s2->hour);
        Serial.printf("min  =%d %d\n", s1->min, s2->min);
    
    if ((s1->year < s2->year) ||
        (s1->year == s2->year && s1->month < s2->month) ||
        (s1->year == s2->year && s1->month == s2->month && s1->day < s2->day) ||
        (s1->year == s2->year && s1->month == s2->month && s1->day == s2->day && s1->hour < s2->hour) ||
        (s1->year == s2->year && s1->month == s2->month && s1->day == s2->day && s1->hour == s2->hour && s1->min < s2->min)){
        return 1;
    } else {
        return 0;
    }
}

void LOG_1(char *s){
    if (FileLog) FileLog.println(s);
}


void loop(){
    int ret;
    
    if (RecvMode == RECVMODE_LISTEN){
        DispMainMenu(SelectedMenu, false);
        ret = PauseAndSelectItem();
        if (ret == SWSELECT_PRESSED){
            if (strcmp(StList[SelectedMenu].id, ID_REPLAY) != 0){
                DispMainMenu(SelectedMenu, true);
                if (Recieve(StList[SelectedMenu].id, 0) == SWPOFF_PRESSED){
                    SaveVolume();
                    Poff();
                }
                SaveVolume();
            } else {
                if (ReplayMenu() == SWPOFF_PRESSED){
                    Poff();
                }
            }
        } else if (ret == SWRETURN_PRESSED){
            // do nothing
        } else if (ret == SWUP_PRESSED){
            if (SelectedMenu > 0){
                SelectedMenu--;
            } else {
                SelectedMenu = NumStList - 1;
            }
        } else if (ret == SWDOWN_PRESSED){
            if (SelectedMenu < (NumStList - 1)){
                SelectedMenu++;
            } else {
                SelectedMenu = 0;
            }
        } else if (ret == SWPOFF_PRESSED){
            Poff();
        }
    } else {
        led_blink_stop();
        ExecScheduledRecording();
        Poff();
    }
}

/*
 *  電源遮断
 */
void Poff(){
    LCD_Println(0, "POFF    ");
    RTC_SetNtpTime();
    LoadSchedule();
    WiFi.disconnect(true);
    WiFi.mode(WIFI_OFF);
    Serial.println("Power off");
    if (FileLog) FileLog.close();
    // 電源遮断
    PowerDown();
//    digitalWrite(POWER_ON, LOW);
//    while (1){
//        delay(1000);
//    }
}

// 電源遮断
void PowerDown(){
    digitalWrite(POWER_ON, LOW);
    while (1){
        delay(1000);
    }
}


/*
 *  プッシュボタンの入力待ち
 */
int PauseAndSelectItem(){
    unsigned long   start_tm;
    start_tm = millis();
    
    do {
        if (EVENT_SWPOFF == true){
            Serial.print("SWPOFF \n");
            EVENT_SWPOFF = false;
            return SWPOFF_PRESSED;
        } else if (EVENT_SWSELECT == true){
            Serial.print("SWSELECT \n");
            EVENT_SWSELECT = false;
            return SWSELECT_PRESSED;
        } else if (EVENT_SWRETURN == true){
            Serial.print("SWRETURN \n");
            EVENT_SWRETURN = false;
            return SWRETURN_PRESSED;
        } else if (EVENT_SWUP == true){
            Serial.print("SWUP \n");
            EVENT_SWUP = false;
            return SWUP_PRESSED;
        } else if (EVENT_SWDOWN == true){
            Serial.print("SWDOWN \n");
            EVENT_SWDOWN = false;
            return SWDOWN_PRESSED;
        } else if (EVENT_SWOPTION == true){
            Serial.print("SWOPTION \n");
            EVENT_SWOPTION = false;
            return SWOPTION_PRESSED;
        } else if ((millis() - start_tm) > AUTO_POFF_TIME){
            EVENT_SWPOFF = false;
            return SWPOFF_PRESSED;
        }
    } while (1);
    
    return 0;
}

/*
 *  メインメニューの文字列表示
 *      at_recieve: trueのとき受信中の局名表示
 */
void DispMainMenu(int selectedmenu, bool at_recieve){
    if (at_recieve == false){
        LCD_Println(0, StList[selectedmenu].name);
        LCD_Println(1, StList[selectedmenu].name + 8);
    } else {
        LCD_Println(0, StList[selectedmenu].name);
        LCD_Println(1, LABEL_RECIEVE);
    }
}

/*
 *  受信
 *      id:ステーションID
 *      save_len_min:録音時間（分） 0のときは無制限（録音しないとき）
 */
int Recieve(char *id, int save_len_min){
    char    *c0, *c1, *c2, *c3, *c4;
    int     found = 0;
    char    name[24];
    char    Url[URL_SIZE];
    char    format[8];
    char    method[16];
    char    outfile[16];
    int     retval = 0;
    //unsigned long filesize;
    
    led_blink_stop();
    
    // ステーションIDに対応するURLを取得
    FileObj = SD.open(STATIONLIST_FILE);
    if(!FileObj){
        Serial.println("Failed to open file for reading");
        return retval;
    }
    Serial.printf("URL ID %s\n", id);
    while (ReadLine(FileObj, FileBuf, URL_SIZE) != 0){
        if (FileBuf[0] == '#') continue;
        
        c0 = FileBuf;
        c1 = strchr(c0, ',');
        if (c1 != NULL){
            *c1 = '\0';
            c1++;
        } else continue;
        c2 = strchr(c1, ',');
        if (c2 != NULL){
            *c2 = '\0';
            c2++;
        } else continue;
        c3 = strchr(c2, ',');
        if (c3 != NULL){
            *c3 = '\0';
            c3++;
        } else continue;
        c4 = strchr(c3, ',');
        if (c4 != NULL){
            *c4 = '\0';
            c4++;
        } else continue;
        
        if (strcmp(c0, id) == 0){
            found = 1;
            strcpy(name, c1);
            strcpy(format, c2);
            strcpy(method, c3);
            strcpy(Url, c4);
            break;
        }
    }
    FileObj.close();
    
    if (found == 0){    // URL不明
        LCD_Println(0, "NO URL  ");
        Serial.println("URL not found");
        delay(5000);
        return retval;
    }
    
    // オーディオアンプ　オン
    if (RecvMode == RECVMODE_LISTEN){
        digitalWrite(AMP_CE, HIGH);
    }
    // 出力ファイル　オープン
    if (GetOutfileName(format, outfile) == true){
        FileSnd = SD.open(outfile, FILE_WRITE);
        LOG_1(outfile);
        if (!FileSnd){
            LOG_1("SND OUTPUT FILE OPEN ERROR");
        } else {
            Serial.printf("open \n");
        }
        RecFileBytes = 0;
        LOG_1(outfile);
    } else {
        LOG_1("CANNOT MAKE OUTPUT FILE NAME");
    }
    
    // 配信方式による大別
    if (strcmp(method, METHOD_M3U) == 0){
        if (recieve_m3u(Url, save_len_min) == SWPOFF_PRESSED){
            retval = SWPOFF_PRESSED;
        }
    } else if (strcmp(method, METHOD_HIGH) == 0){
        if (RecvMode == RECVMODE_LISTEN){
            if (recieve_streaming_highbitrate(Url) == SWPOFF_PRESSED){
                retval = SWPOFF_PRESSED;
            }
        } else {
            if (recieve_streaming_schrec(Url, save_len_min) == SWPOFF_PRESSED){
                retval = SWPOFF_PRESSED;
            }
        }
    } else if (strcmp(method, METHOD_NORMAL) == 0){
        if (RecvMode == RECVMODE_LISTEN){
            if (recieve_streaming_listen(Url) == SWPOFF_PRESSED){
                retval = SWPOFF_PRESSED;
            }
        } else {
            if (recieve_streaming_schrec(Url, save_len_min) == SWPOFF_PRESSED){
                retval = SWPOFF_PRESSED;
            }
        }
    }
    
    // オーディオアンプ　オフ
    if (RecvMode == RECVMODE_LISTEN){
        digitalWrite(AMP_CE, LOW);
    }
    // 出力ファイル　クローズ
    if (FileSnd){
        //filesize = FileSnd.size();    // ファイル書き込みされなかったとき ０にならない。
        FileSnd.close();
        Serial.printf("Recfile size=%ld \n", RecFileBytes);
        LOG_1("SND OUTPUT FILE CLOSE");
        
        if (RecFileBytes == 0){
            SD.remove(outfile);
            Serial.println("Recfile removed");
            LOG_1("Recfile removed");
        }
    }
    
    return retval;
}

/*
 *  現在日時
 */
void DispCurrentDateTime(){
    uint8_t     dt[6];
    char        dts[20];
    RTC_GetCurrentDateTime(dt);
    sprintf(dts, "%02d/%02d/%02d %02d:%02d", dt[0], dt[1], dt[2], dt[3], dt[4]);
    if (FileLog) FileLog.println(dts);
    Serial.printf("%s\n", dts);
}

/*
 *  出力ファイル名を決める
 */
bool GetOutfileName(char *extention, char *outfile){
    uint8_t     dt[6];
    int         i;
    
    RTC_GetCurrentDateTime(dt);
    
    for (i=1; i<=99; i++){
        sprintf(outfile, "/%02d%02d%02d%02d.%s", dt[0], dt[1], dt[2], i, "MP3");
        if (SD.exists(outfile) == true){
            continue;
        }
        sprintf(outfile, "/%02d%02d%02d%02d.%s", dt[0], dt[1], dt[2], i, "AAC");
        if (SD.exists(outfile) == true){
            continue;
        }
        sprintf(outfile, "/%02d%02d%02d%02d.%s", dt[0], dt[1], dt[2], i, extention);
        return true;
    }
    return false;
}

/*
 *  プログラム録音
 */
int ExecScheduledRecording(){
    SCHEDULE    w;
    
    memset(&w, 0, sizeof(SCHEDULE));
    FileObj = SD.open(NEXT_SCHEDULE_FILE);
    if(!FileObj){
        Serial.println("NoSchdule");
        return 0;
    }
    if (ReadLine(FileObj, FileBuf, URL_SIZE) != 0){
        sscanf(FileBuf, "%s %d %d %d %d %d %d", w.sta, &w.year, &w.month, &w.day, &w.hour, &w.min, &w.len);
        FileObj.close();
    } else {
        FileObj.close();
        return 0;
    }
    
    if (w.sta[0] == '\0' || w.len == 0){
        return 0;
    }
    
    Recieve(w.sta, w.len);
    
    return 0;
}

/*
 *    録音データファイルの選択画面表示
 */
int ReplayMenu(){
    int ret;
    char    buf[24];
    char    *ext;
    
    if (LoadRecFiles() == 0) return 0;
    
    SelectedFile = 0;
    while (1){
        LCD_Println(0, RecFileList[SelectedFile].file);     //
        ext = strchr(RecFileList[SelectedFile].file, '.');
        if (RecFileList[SelectedFile].min == 0){
            buf[0] = '/';
            strcpy(buf + 1, RecFileList[SelectedFile].file);
            RecFileList[SelectedFile].min = get_minutes_sound(buf, RecFileList[SelectedFile].size);
        }
        sprintf(buf, "%3d %3d%s", RecFileList[SelectedFile].size / 1048576, RecFileList[SelectedFile].min, ext + 1);
        //sprintf(buf, "%dMB %s  ", RecFileList[SelectedFile].size / 1048576, ext + 1);
        LCD_Println(1, buf);
        
        ret = PauseAndSelectItem();
        if (ret == SWSELECT_PRESSED){
            LCD_Println(1, LABEL_REPLAY);
            if (Replay(RecFileList[SelectedFile].file) == SWPOFF_PRESSED){
                return SWPOFF_PRESSED;
            }
            LCD_Println(1, LABEL_CLEAR);
        } else if (ret == SWOPTION_PRESSED){
            if (ReplayOptionMenu(RecFileList[SelectedFile].file, RecFileList[SelectedFile].size) == SWPOFF_PRESSED){
                return SWPOFF_PRESSED;
            }
            if (LoadRecFiles() == 0) return 0;
            SelectedFile = 0;
        } else if (ret == SWRETURN_PRESSED){
            return 0;
        } else if (ret == SWUP_PRESSED){
            if (SelectedFile > 0){
                SelectedFile--;
            } else {
                SelectedFile = NumRecFileList - 1;
            }
        } else if (ret == SWDOWN_PRESSED){
            if (SelectedFile < (NumRecFileList - 1)){
                SelectedFile++;
            } else {
                SelectedFile = 0;
            }
        } else if (ret == SWPOFF_PRESSED){
            return SWPOFF_PRESSED;
        }
    }
    
    return 0;
}

/*
 *    オプションメニュー
 *        file:ファイルパス、size:ファイルサイズ（バイト）
 */
int ReplayOptionMenu(char *file, size_t size){
    int sel_opt = 0;
    int num_items = 3;
    int ret;
    int minutes = 0;
    char buf[16];
    char *ext = strchr(file, '.');
    
    LCD_Println(0, file);
    
    while (1){
        switch (sel_opt){
        case 0: sprintf(buf, "%dM %s  ", size / 1048576, ext + 1); break;
        case 1: sprintf(buf, " Delete "); break;
        case 2: sprintf(buf, " %d min ", minutes); break;
        }
        LCD_Println(1, buf); 
        ret = PauseAndSelectItem();
        if (ret == SWSELECT_PRESSED){
            if (sel_opt == 1){            // Delete
                buf[0] = '/';
                strcpy(buf + 1, file);
                SD.remove(buf);
                break;
            } else if (sel_opt == 2){
                buf[0] = '/';
                strcpy(buf + 1, file);
                minutes = get_minutes_sound(buf, size);
            }
        } else if (ret == SWRETURN_PRESSED){
            break;
        } else if (ret == SWUP_PRESSED){
            if (sel_opt > 0){
                sel_opt--;
            } else {
                sel_opt = num_items - 1;
            }
        } else if (ret == SWDOWN_PRESSED){
            sel_opt++;
            if (sel_opt == num_items) sel_opt = 0;
        } else if (ret == SWPOFF_PRESSED){
            return SWPOFF_PRESSED;
        }
    }
    
    return 0;
}

// 音源ファイルの時間長さ（分）
int get_minutes_sound(char *filepath, size_t size){
    char    *c1, *c2;
    
    c1 = strchr(filepath, '.');
    if (c1 == NULL) return 0;
    
    if (strcmp(c1, ".MP3") == 0){
        return get_minutes_MP3(filepath, size);
    } else if (strcmp(c1, ".AAC") == 0){
        return get_minutes_AAC(filepath, size);
    }
    
    return 0;
}

/*
 *  MP3ファイルの再生時間長を分で取得
 *  ファイルを最後までなめると時間がかかるので概算で。
 *      size:ファイルサイズ（バイト）
 */
int get_minutes_MP3(char *filepath, size_t size){
    int     i, get_size;
    int     bitrate_index;
    int     sample_index;
    int     padding;
    int     frame_len;
    long    sum_frame_len;
    long    frames = 0;
    int     databytes;
    int     avg_frames = 500;      // 概算するためのフレーム数
    int     length = 0;
    
    FileSnd = SD.open(filepath);
    if(!FileSnd){
        Serial.println("Failed to open file for reading");
        return 0;
    }
    
    sum_frame_len = 0;
    get_size = FileSnd.read(RecvBuff, RECVBUFF_SIZE);
    databytes = get_size;       // RecvBuffの有効バイト数
    if (databytes < MP3_HDR_BYTE){
        FileSnd.close();
        return 0;
    }
    i = 0;
        
    while (1){
        if (i < (databytes - MP3_HDR_BYTE)){
            // フレームヘッダを探す
            for (; i<(databytes - MP3_HDR_BYTE); i++){
                if (RecvBuff[i] == 0xFF && (RecvBuff[i + 1] == 0xFB || RecvBuff[i + 1] == 0xFA)){
                    break;
                }
            }
            //Serial.printf(" header start=%d\n", i);
            
            // フレームヘッダあり
            if (i < (databytes - MP3_HDR_BYTE)){
                bitrate_index = (RecvBuff[i + 2] >> 4);
                sample_index = ((RecvBuff[i + 2] & 0x0C) >> 2);
                padding = ((RecvBuff[i + 2] & 0x02) >> 1);
                frame_len = int((144 * bitrate_tab[bitrate_index] * 1000 / sample_tab[sample_index]) + padding);    // フレームサイズ（ヘッダのサイズ４バイトを含む）
                if (frame_len > MP3_HDR_BYTE){
                    if ((i + frame_len) < (databytes - MP3_HDR_BYTE)){
                        int i2 = i + frame_len;      // 次のフレームヘッダのはず
                        if (RecvBuff[i2] == 0xFF && (RecvBuff[i2 + 1] == 0xFB || RecvBuff[i2 + 1] == 0xFA)){
                            i = i2;
                            frames++;
                            sum_frame_len += frame_len;
                            if (frames == avg_frames){      // 平均で計算するためループを抜ける
                                Serial.printf(" frames=%d\n", frames);
                                Serial.printf(" sum_frame_len=%d\n", sum_frame_len);
                                if (sum_frame_len > 0 && (sample_index >= 0 && sample_index <= 3)) {
                                    long e_frames;      // 見積フレーム総数
                                    //e_frames = size / (sum_frame_len / avg_frames);     // ファイルサイズを平均フレームサイズでわることでフレーム総数を概算
                                    //e_frames = ((long)size * avg_frames) / sum_frame_len;     // オーバーフローする？
                                    //Serial.printf(" e_frames=%d\n", e_frames);
                                    e_frames = size / (sum_frame_len / avg_frames);
                                    Serial.printf(" e_frames=%d\n", e_frames);
                                    length = ((e_frames * 1152) / sample_tab[sample_index]) / 60;
                                    Serial.printf(" length=%d\n",  length);
                                }
                                break;
                            }
                            //Serial.printf(" frames=%d\n", frames);
                        } else {    // 次のフレームにつながっていない
                            i++;
                            Serial.printf(" No next frame\n");
                        }
                    } else {
                        memcpy(RecvBuff, RecvBuff + i, databytes - i);
                        get_size = FileSnd.read(RecvBuff + databytes - i, RECVBUFF_SIZE - (databytes - i));
                        if (get_size == 0){
                            break;
                        }
                        databytes = databytes - i + get_size;
                        if (databytes < MP3_HDR_BYTE){
                            break;
                        }
                        //Serial.printf(" exceed frame\n");
                        i = 0;
                    }
                } else {    // エラー
                    i++;    // 再度フレームヘッダを探す
                    Serial.printf(" invalid frame len=%d\n", frame_len);
                }
            } else {    // フレームヘッダない
                memcpy(RecvBuff, RecvBuff + databytes - MP3_HDR_BYTE, databytes - MP3_HDR_BYTE);
                get_size = FileSnd.read(RecvBuff + databytes - MP3_HDR_BYTE, RECVBUFF_SIZE - (databytes - MP3_HDR_BYTE));
                if (get_size == 0){
                    break;
                }
                databytes = databytes - MP3_HDR_BYTE + get_size;
                if (databytes < MP3_HDR_BYTE){
                    break;
                }
                i = 0;
                Serial.printf(" No frame\n");
            }
        } else if (i < databytes){
            memcpy(RecvBuff, RecvBuff + i, databytes - i);
            get_size = FileSnd.read(RecvBuff + databytes - i, RECVBUFF_SIZE - (databytes - i));
            if (get_size == 0){
                break;
            }
            databytes = databytes - i + get_size;
            if (databytes < MP3_HDR_BYTE){
                break;
            }
            i = 0;
            Serial.printf(" 1 \n");
        } else {
            get_size = FileSnd.read(RecvBuff, RECVBUFF_SIZE);
            if (get_size == 0){
                break;
            }
            databytes = get_size;
            if (databytes < MP3_HDR_BYTE){
                break;
            }
            i = 0;
            Serial.printf(" 2 \n");
        }
        
    }
    FileSnd.close();
    
#if 0       // フレームをフルカウントして時間長さを得る
    Serial.printf(" frames=%d\n", frames);
    Serial.printf(" sum_frame_len=%d\n", sum_frame_len);
    if (sum_frame_len > 0 && (sample_index >= 0 && sample_index <= 3)) {
        Serial.printf(" length=%ld\n",  ((frames * 1152) / sample_tab[sample_index]) / 60);
        //Serial.printf(" length=%ld\n",  (frames * ((double)1152 / (double)sample_tab[sample_index])) / 60);
    }
#endif
    
    return length;
}


/*
 *  AACファイルの再生時間長を分で取得
 *  ファイルを最後までなめると時間がかかるので概算で
 *      size:ファイルサイズ（バイト）
 */
int get_minutes_AAC(char *filepath, size_t size){
    int     i, get_size;
    int     minutes = 0;
    int     frame_len;
    long    duration = 0;
    long    frames = 0;
    long    sum_frame_len = 0;
    int     sampl_f;
    int     databytes;
    int     avg_frames = 500;      // 概算するためのフレーム数
    int     length = 0;
    
    FileSnd = SD.open(filepath);
    if(!FileSnd){
        Serial.println("Failed to open file for reading");
        return 0;
    }
    get_size = FileSnd.read(RecvBuff, RECVBUFF_SIZE);
    databytes = get_size;       // RecvBuffの有効バイト数
    if (databytes < AAC_HDR_BYTE){
        FileSnd.close();
        return 0;
    }
    i = 0;
    
    while (1){
        if (i < databytes - AAC_HDR_BYTE){
            // ADTSヘッダを探す
            for (; i<(databytes - AAC_HDR_BYTE); i++){
                if (RecvBuff[i] == 0xFF && (RecvBuff[i + 1] & 0xF0) == 0xF0) {
                    break;
                }
            }
        }
        
        // ADTSヘッダあり
        if (i < (databytes - AAC_HDR_BYTE)){
            int sampl = ((RecvBuff[i + 2] >> 2) & 0x0F);
            if (sampl >= 0 && sampl < NumAdtsSampling) {
                sampl_f = AdtsSampling[sampl];
            } else {
                sampl_f = 0xFFFF;
            }

            frame_len = (((int)RecvBuff[i + 3] & 0x03) << 11) | (((int)RecvBuff[i + 4]) << 3) | ((RecvBuff[i + 5] >> 5) & 0x07);
            if ((i + frame_len) < (databytes - AAC_HDR_BYTE)){
                if (sampl_f != 0xFFFF) {
                    frames++;
                    sum_frame_len += frame_len;
                    duration += (1024*1000) / sampl_f;          // 1024 or 1000 ?
                    if (frames == avg_frames){
                        length = (duration * (size / sum_frame_len)) / (60*1000);
                        Serial.printf("duration=%ld\n", duration);
                        Serial.printf("sum_frame_len=%ld\n", sum_frame_len);
                        Serial.printf("length=%d\n", length);
                        break;
                    }
                }
                i += frame_len;
            } else {
                memcpy(RecvBuff, RecvBuff + i, databytes - i);
                get_size = FileSnd.read(RecvBuff + databytes - i, RECVBUFF_SIZE - (databytes - i));
                if (get_size == 0){
                    break;
                }
                databytes = databytes - i + get_size;
                if (databytes < AAC_HDR_BYTE){
                    break;
                }
                //Serial.printf(" exceed frame\n");
                i = 0;
            }
        } else if (i < databytes){  // ヘッダない
            memcpy(RecvBuff, RecvBuff + i, databytes - i);
            get_size = FileSnd.read(RecvBuff + databytes - i, RECVBUFF_SIZE - (databytes - i));
            if (get_size == 0){
                break;
            }
            databytes = databytes - i + get_size;
            if (databytes < AAC_HDR_BYTE){
                break;
            }
            i = 0;
            Serial.printf(" 1 \n");
        } else {    // ヘッダない
            get_size = FileSnd.read(RecvBuff, RECVBUFF_SIZE);
            if (get_size == 0){
                break;
            }
            databytes = get_size;
            if (databytes < AAC_HDR_BYTE){
                break;
            }
            i = 0;
            Serial.printf(" No frame\n");
        }
    }
    FileSnd.close();
    
#if 0
    Serial.printf("frames=%ld\n", frames);
    Serial.printf("sum_frame_len=%ld\n", sum_frame_len);
    Serial.printf("duration=%ld\n", duration / (60*1000));
    return (duration / (60*1000));
#endif
    
    return length;
}



/*
 *  録音データファイルのリストを作成
 */
int LoadRecFiles(){
    FileObj = SD.open("/");
    NumRecFileList = 0;
    
    while (1){
        File    entry = FileObj.openNextFile();
        if (!entry){
            break;
        }
        if (!entry.isDirectory()){      // ファイル
            char    *c1, *c2;
            c1 = (char *)entry.name();
            if (c1 != NULL){
                //Serial.printf("file=%s\n", c1);
                c2 = strchr(c1, '.');
                if (c2 != NULL){
                    if ((c2 - c1) <= 11 && (strcmp(c2, ".MP3") == 0 || strcmp(c2, ".AAC") == 0)){    // "/yymmddnn.MP3", "/yymmddnn.AAC"
                        if (*c1 == '/'){
                            strcpy(RecFileList[NumRecFileList].file, c1 + 1);
                        } else {
                            strcpy(RecFileList[NumRecFileList].file, c1);
                        }
                        RecFileList[NumRecFileList].size = entry.size();
                        RecFileList[NumRecFileList].min = 0;
                        NumRecFileList++;
                        Serial.printf(" rec file=%s\n", c1);
                        if (NumRecFileList == RECFILELIST_SIZE) break;
                    }
                }
            }
        }
    }
    FileObj.close();
    
    SortRecFileList(RecFileList, 0, NumRecFileList - 1);
    
    return NumRecFileList;
}

/*
 *    nameの大きい順(新しい順)にソート
 */
void SortRecFileList(REPLAY_FILE arr[], int left, int right) {
    int i = left, j = right;
    REPLAY_FILE tmp;
    REPLAY_FILE pivot = arr[(left + right) / 2];

    /* partition */
    while (i <= j) {
        while (strcmp(arr[i].file, pivot.file) > 0)
            i++;
        
        while (strcmp(arr[j].file, pivot.file) < 0)
            j--;
        if (i <= j) {
            tmp = arr[i];
            arr[i] = arr[j];
            arr[j] = tmp;
            i++;
            j--;
        }
    };

    /* recursion */
    if (left < j)
        SortRecFileList(arr, left, j);
    if (i < right)
        SortRecFileList(arr, i, right);
}

/*
 *  保存ファイルの再生
 */
int Replay(char *file){
    int         emp_size;
    int         get_size;
    int         ret = 0;
    char        filepath[FILEPATH_SIZE];
    
    Serial.print("Replay \n");
    Decoding = false;
    
    filepath[0] = '/';
    strcpy(filepath + 1, file);
    
    FileSnd = SD.open(filepath);
    if(!FileSnd){
        Serial.println("Failed to open file for reading");
        return ret;
    }
    if (FileSnd.read(RecvBuff, RECVBUFF_SIZE) == 0){
        FileSnd.close();
        return ret;
    }
    dat_start = 0;
    dat_size = 0;
    emp_size = 0;
    emp_start = 0;
    
    digitalWrite(AMP_CE, HIGH);
    EnableDreqInt = true;       // DREQの立ち上がりでDecode()を開始
    
    while (1){
        if (digitalRead(MP3_DREQ) != 0 /*&& dat_size > 2000*/){
            //Serial.printf("dat_size1=%d \n", dat_size);
            //if (Decoding == false){     // isr_dreq()からDecode()を実行中でないこと （このif不要 2022/01/16)
                detachInterrupt(MP3_DREQ);
                Decode();
                attachInterrupt(MP3_DREQ, isr_dreq, RISING);
            //}
        }
        
        //EnableDreqInt = true;       // DREQの立ち上がりでDecode()を開始
        emp_size = RECVBUFF_SIZE - dat_size;
        if (emp_start + emp_size < RECVBUFF_SIZE){
            get_size = emp_size;
        } else {
            get_size = RECVBUFF_SIZE - emp_start;
        }
        if (get_size > 0){
            //Serial.printf("dreq=%d get_size=%d\n", digitalRead(MP3_DREQ), get_size);
            if (FileSnd.read(RecvBuff + emp_start, get_size) == 0) break;
            emp_start += get_size;
            if (emp_start == RECVBUFF_SIZE){
                emp_start = 0;
            }
            dat_size += get_size;
        } else {
            //Serial.printf("dreq=%d get_size=%d\n", digitalRead(MP3_DREQ), get_size);
        }
        //EnableDreqInt = false;
        
        if (EVENT_SWPOFF == true){
            ret = SWPOFF_PRESSED;
            break;
        } else if (EVENT_SWRETURN == true){
            break;
        }
    }
    
    Serial.print("Replay end \n");
    digitalWrite(AMP_CE, LOW);
    FileSnd.close();
    EnableDreqInt = false;
    
    SaveVolume();
    return ret;
}


/*
 *  M3U方式のステーションを受信
 */
int recieve_m3u(char *Url, int save_len_min){
    int         httpCode;
    int         data_read;
    int         get_size;
    int         emp_size;
    int         sum;
    int         retry_sum;
    uint8_t     recvbuff[M3UBUFF_SIZE];
    uint8_t     skipbuff[SKIPBUFF_SIZE];
    int         m3u_size;   // M3Uファイルサイズ
    char        fileurl[URL_SIZE];
    char        *c1;
    int         i;
    int         sta_type;
    char        m3u_nl;         // M3Uファイルの改行コード（２バイトの時は先頭バイト）
    int         m3u_nllen;      // 同 バイト数
    int         continue_flag;
    unsigned long   start_tm;
    unsigned long   save_len_ms;
    unsigned long   get_sound_tm;
    unsigned long    start_delay_ms;
    bool        retry_flag;
    int         retry_cnt;
    int         retry_get_cnt;
    bool        rcv_rec_flag = false;       // 受信しながら録音
    int         ret;
    int         retval = 0;
    
    // put your main code here, to run repeatedly:
    // wait for WiFi connection
    retry_cnt = 0;
  
    HTTPClient http;
  
    Serial.print("[HTTP] begin...\n");
    
    if (strstr(Url, ".nhk.jp") != NULL){
        sta_type = M3U_NHK;
        m3u_nl = 0x0a;
        m3u_nllen = 1;
        start_delay_ms = SCHREC_POWERUP_MARGIN*60*1000 + (1*60 + 30)*1000;     // 実測による値
    } else if (strstr(Url, "musicbird") != NULL){
        sta_type = M3U_MUSICBIRD;
        m3u_nl = 0x0d;
        m3u_nllen = 2;
        start_delay_ms = SCHREC_POWERUP_MARGIN*60*1000;
    } else {
        LCD_Println(0, "Invalid ");
        LCD_Println(1, "URL     ");
        LOG_1("Invalid URL");
        LOG_1(Url);
        delay(5000);
        return retval;
    }
    
    NumM3uTable = 0;
    M3uTableInsIndex = 0;
    M3uTableGetIndex = 0;
    EnableDreqInt = true;
    Decoding = false;
    
    if (sta_type == M3U_MUSICBIRD){
        /*
         *  局リストのURLから取得できるM3Uファイルには別URLが記載されている。音源取得はそのURLで。
         * M3Uファイルの例：
        #EXTM3U
        #EXT-X-STREAM-INF:BANDWIDTH=56000,CODECS="mp4a.40.5"
        https://chi-live3-ais-relay1.leanstream.co/musicbird/JCB048.stream/playlist.m3u8?listeningSessionID=61bfec7700b942b8_1116098_idY9aERM_NjkuNjUuNDYuMzY6ODAwMA!!_00000003O6T&downloadSessionID=0
         */
        // 
        ret = GetM3U(http, Url, (char *)recvbuff, M3UBUFF_SIZE, &m3u_size);
        if (ret != M3U_GET_OK){
            return retval;
        }
        // M3Uファイル解析
        ParseM3U((char *)recvbuff, m3u_size, m3u_nl, m3u_nllen);
        if (NumM3uTable > 0){
            // https:をhttp:に置き換える
            c1 = strchr(M3uTable[0].reqfile, ':');
            strcpy(Url, "http");
            strcpy(Url + 4, c1);
            //strcpy(Url, M3uTable[0].reqfile);
        } else {
            LOG_1("Invalid M3U file");
            return retval;
        }
        NumM3uTable = 0;
        M3uTableInsIndex = 0;
    }
    
    // プログラム録音時、録音開始までの遅延を入れる
    if (RecvMode == RECVMODE_SCHREC){
        if (start_delay_ms > 0){
            wait_ms(start_delay_ms);
            //start_tm = millis(start_delay_ms);
            //do {
            //    delay(10);
            //} while ((millis() - start_tm) < start_delay_ms);
        }
    }
    
    save_len_ms = save_len_min * 60 * 1000;       // 録音時間ミリ秒単位
    start_tm = millis();
    
    continue_flag = 1;
    retry_cnt = 0;
    while (continue_flag){
        // M3UファイルURLをGET
        ret = GetM3U(http, Url, (char *)recvbuff, M3UBUFF_SIZE, &m3u_size);
        if (ret != M3U_GET_OK){
            break;
        }
        // M3Uファイル解析
        ParseM3U((char *)recvbuff, m3u_size, m3u_nl, m3u_nllen);
        
//        if (DreqEvent == true){
//            Serial.printf("DREQ Event 1\n");
//            DreqEvent = false;
//            if (Decoding == true){
//                Serial.printf("  Decoding\n");
//            }
//        }
        
        if (NumM3uTable == 0){
            unsigned long   tm1;
            
            Serial.printf("NO m3u\n");
            tm1 = millis();
            while ((millis() - tm1) < 1000){
                delay(1);
            }
//            if (DreqEvent == true){
//                Serial.printf("DREQ Event 2\n");
//                DreqEvent = false;
//                if (Decoding == true){
//                    Serial.printf("  Decoding\n");
//                }
//            }
            continue;
        }
        
        // get sound
        strcpy(fileurl, Url);
        char    *c1 = strrchr(fileurl, '/');
        
        while (NumM3uTable > 0 && continue_flag == 1){
            if (sta_type == M3U_NHK){
                // M3UファイルのURLとファイルパスから音源のURLを作る
                strcpy(c1 + 1, M3uTable[M3uTableGetIndex].reqfile);
            } else if (sta_type == M3U_MUSICBIRD){
                // https:をhttp:に置き換える
                c1 = strchr(M3uTable[M3uTableGetIndex].reqfile, ':');
                strcpy(fileurl, "http");
                strcpy(fileurl + 4, c1);
            }
            Serial.printf("[File Get] : %d %s\n", M3uTableGetIndex, fileurl);
            // 音源URLをGET
            retry_flag = false;
            retry_cnt = 0;
            retry_get_cnt = 0;
            
            do {    // 音源ファイル取得・リトライループ
                int     len;
                
                get_sound_tm = millis();    // Timeoutの起点
                http.begin(fileurl);
                httpCode = http.GET();
                if (httpCode != HTTP_CODE_OK /*200*/){
                    char    msg[32];
                    sprintf(msg, "RETRY GET SND (%d/%d)", httpCode, retry_get_cnt);
                    LOG_1(msg);
                    LOG_1(fileurl);
                    Serial.printf("[HTTP] GET... code: %d (%s)\n", httpCode, fileurl);
                    http.end();
                    if (retry_get_cnt >= MAX_RETRY_COUNT){
                        digitalWrite(POWER_LED, HIGH);
                        return retval;           // 受信不可能
                    } else {
                        retry_get_cnt++;
                        continue;
                    }
                }
                
                len = http.getSize();   // Content-Length
                WiFiClient * stream = http.getStreamPtr();
                sum = 0;
                
                Serial.printf("[HTTP] GET... size: %d\n", len);
                
                if (RecvMode == RECVMODE_LISTEN){
                    unsigned long   tm1 = millis();
                    unsigned long    tm_sum = 0;        // available() から readBytes() までの実行時間合計
                    int             get_cnt = 0;        // readBytes()の回数
                    int                cnt0 = 0;        // available()==0の回数
                    while (sum < len){
                        tm1 = millis();
                        size_t size = stream->available();
                        if (size > 0 /*&& dat_size < RECVBUFF_SIZE*/) {
                            if (retry_flag == false && dat_size < RECVBUFF_SIZE){
                                emp_size = RECVBUFF_SIZE - dat_size;
                                if (emp_start + emp_size < RECVBUFF_SIZE){
                                    get_size = emp_size;
                                } else {
                                    get_size = RECVBUFF_SIZE - emp_start;
                                }
                                if (get_size > size) get_size = size;
                                data_read = stream->readBytes(RecvBuff + emp_start, get_size);
                                if (data_read > 0){
                                    emp_start += data_read;
                                    if (emp_start == RECVBUFF_SIZE){
                                        emp_start = 0;
                                        // ファイル出力
                                        if (rcv_rec_flag == true){
                                            FileSnd.write(RecvBuff, RECVBUFF_SIZE);
                                            FileSnd.flush();
                                            RecFileBytes += RECVBUFF_SIZE;
                                        }
                                    }
                                    dat_size += data_read;
                                    sum += data_read;
                                }
                                tm_sum += (millis() - tm1);        // 音源データを取得するための実所要時間を取得
                                get_cnt++;
                            } else if (retry_flag == true){
                                if (retry_sum > 0){
                                    int remain = retry_sum - sum;
                                    if (remain > SKIPBUFF_SIZE){
                                        get_size = SKIPBUFF_SIZE;
                                    } else {
                                        get_size = remain;
                                    }
                                    data_read = stream->readBytes(skipbuff, get_size);
                                    sum += data_read;
                                    if (sum >= retry_sum){
                                        retry_flag = false;
                                    }
                                } else {
                                    retry_flag = false;
                                }
                            }
                        } else if (size == 0){
                            cnt0++;
                            if ((millis() - get_sound_tm) > GETSOUND_TIMEOUT){
                                Serial.printf("timeout sum=%d retry=%d\n", sum, retry_cnt);
                                if (retry_cnt == 0 || sum > retry_sum){
                                    retry_sum = sum;
                                }
                                retry_cnt++;
                                retry_flag = true;
                                break;
                            }
                        }
                        if (digitalRead(MP3_DREQ) != 0){
                            detachInterrupt(MP3_DREQ);
                            Decode();
                            attachInterrupt(MP3_DREQ, isr_dreq, RISING);
                        }
                        
                        // SW
                        if (EVENT_SWRETURN == true){
                            Serial.printf("SWRETURN\n");
                            EVENT_SWRETURN = false;
                            continue_flag = 0;
                            break;
                        } else if (EVENT_SWPOFF == true){
                            retval = SWPOFF_PRESSED;
                            EVENT_SWPOFF = false;
                            continue_flag = 0;
                            break;
                        } else if (EVENT_SWOPTION == true){
                            EVENT_SWOPTION = false;
                            if (rcv_rec_flag == false && FileSnd){
                                rcv_rec_flag = true;
                                LCD_Println(1, LABEL_RECIEVE_RECORDING);
                            } else {
                                rcv_rec_flag = false;
                                LCD_Println(1, LABEL_RECIEVE);
                            }
                        }
                    }
                    
                    // 音源ファイル一つをGETするのに要する正味時間
                    // time : available() ～ readBytes() の所要時間の合計 (1150ms以下)
                    // count : readBytes()の回数 (230程度)
                    // count0 : available()がゼロの回数 (0であることが多いが、最大6000ぐらい)
                    //Serial.printf("count=%d, count0=%d time=%d\n", get_cnt, cnt0, tm_sum);
                } else if (RecvMode == RECVMODE_SCHREC){        // 音声データを保存だけする
                    while (sum < len){
                        size_t size = stream->available();
                        if (size > 0) {
                            if (retry_flag == false /*&& dat_size < RECVBUFF_SIZE*/){
                                if (size > RECVBUFF_SIZE){
                                    get_size = RECVBUFF_SIZE;
                                } else {
                                    get_size = size;
                                }
                                data_read = stream->readBytes(RecvBuff, get_size);
                                if (data_read > 0){
                                    FileSnd.write(RecvBuff, data_read);
                                    FileSnd.flush();
                                    RecFileBytes += data_read;
                                    sum += data_read;
                                    //Serial.printf("data_read=%d sum=%d\n", data_read, sum);
                                }
                            } else if (retry_flag == true){
                                if (retry_sum > 0){
                                    int remain = retry_sum - sum;
                                    if (remain > SKIPBUFF_SIZE){
                                        get_size = SKIPBUFF_SIZE;
                                    } else {
                                        get_size = remain;
                                    }
                                    data_read = stream->readBytes(skipbuff, get_size);
                                    sum += data_read;
                                    if (sum >= retry_sum){
                                        retry_flag = false;
                                    }
                                } else {
                                    retry_flag = false;
                                }
                            }
                        } else {
                            if ((millis() - get_sound_tm) > GETSOUND_TIMEOUT){
                                Serial.printf("timeout\n");
                                if (retry_cnt == 0 || sum > retry_sum){
                                    retry_sum = sum;
                                }
                                retry_cnt++;
                                retry_flag = true;
                                break;
                            }
                        }
                        // SW
                        if (EVENT_SWRETURN == true){
                            Serial.printf("SWRETURN\n");
                            EVENT_SWRETURN = false;
                            continue_flag = 0;
                            break;
                        }
                        // RTC timeout
                        if ((millis() - start_tm) > save_len_ms){
                            continue_flag = 0;
                            break;
                        }
                    }
                }
                
                http.end();
                
                if (retry_flag == true){
                    Serial.printf("retry_get_sound (%d)\n", retry_cnt);
                } else {
                    // 次の音源ファイルへ
                    NumM3uTable--;
                    M3uTableGetIndex++;
                    if (M3uTableGetIndex == M3U_TABLE_SIZE) M3uTableGetIndex = 0;
                }
                
            } while (retry_flag == true && continue_flag == 1);
            
        }   // m3uファイル内・ループ
    }   // m3uファイル取得ループ
    
    EnableDreqInt = false;
    Serial.printf("http.end ...\n");
    http.end();
    return retval;
}

/*
 *  M3UファイルGET
 *  timeoutを考慮していない
 */
int GetM3U(HTTPClient& http, char *Url, char *recvbuff, int buffsize, int *m3u_size){
    int         data_read;
    int         ret;
    int         retry_cnt;
    char        msg[32];
    
    retry_cnt = 0;
    while (1){
        ret = M3U_GET_ERROR;
        *m3u_size = 0;
        http.begin(Url);
        int httpCode = http.GET();
        if (httpCode == HTTP_CODE_OK /*200*/) {
            Serial.printf("[HTTP] GET... code: %d\n", httpCode);
            int len = http.getSize();
            Serial.printf("[HTTP] GET.. length: %d\n", len);
            // M3Uファイル CONTENT
            WiFiClient * stream = http.getStreamPtr();
            if (http.connected()) {
                // get available data size
                while (*m3u_size < len && *m3u_size < M3UBUFF_SIZE){
                    size_t size = stream->available();
                    if (size) {
                        if ((*m3u_size + size ) < buffsize){
                            data_read = stream->readBytes(recvbuff + *m3u_size, size);
                            *m3u_size += data_read;
                            if (*m3u_size == len){
                                ret = M3U_GET_OK;
                                break;
                            }
                        } else {
                            ret = M3U_GET_TOO_LARGE;
                            break;
                        }
                    }
                }
                recvbuff[*m3u_size] = 0;     // ファイルの終端
                Serial.printf("[HTTP] GET.. m3u file: %s\n", recvbuff);
            }
        }
        http.end();
        
        if (ret != M3U_GET_ERROR) break;
        else if (retry_cnt == MAX_RETRY_COUNT_M3U) break;
        
        sprintf(msg, "RETRY GET M3U (%d/%d)", httpCode, retry_cnt);
        LOG_1(msg);
        wait_ms(retry_cnt * RETRY_WAIT_TIME);       // 待ち時間を少しずつ長くする
        retry_cnt++;
    }
    return ret;
}


#if 0
/*
 *    JAZZ24専用：未完
 */
int recieve_jazz24(char *Url, int save_len_min){
    HTTPClient http;
    uint8_t     recvbuff[M3UBUFF_SIZE];
    int         m3u_size;   // M3Uファイルサイズ
    int         i;
    
    NumM3uTable = 0;
    M3uTableInsIndex = 0;
    
    GetM3U(http, Url, (char *)recvbuff, M3UBUFF_SIZE, &m3u_size);
    ParseM3U((char *)recvbuff, m3u_size, 0x0a, 1);
    
    for (i=0; i<NumM3uTable; i++){
        Serial.printf("%d %s\n", i, M3uTable[i].reqfile);
    }
    if (NumM3uTable > 0){
        recieve_streaming_listen(M3uTable[0].reqfile /*, save_len_min*/);    // Transfer-Encoding not supported のエラーになる
    }
    
    return 0;
}
#endif

/*
 *  ストリーミング方式の受信
 */
int recieve_streaming_listen(char *Url){
    int data_read;
    int get_size;
    int emp_size;
    unsigned long   start_tm;
    //unsigned long   save_len_ms;
    unsigned long   save_frames;
    unsigned long   frames;
    int     i;
    int     continue_flag = 0;
    int     reconnect_flag;
    unsigned long   get_sound_tm;
    int     retval = 0;
    int        retry_cnt;
    
    LOG_1("recieve_streaming_listen");
    LOG_1(Url);
    
    SaveAtDecode = false;
    
    //SampleRate = 0;
    //BitRate = 0;
    //save_len_ms = save_len_min * 60 * 1000;       // 録音時間ミリ秒単位
    save_frames = 0;
    frames = 0;
    retry_cnt = 0;
    start_tm = millis();
    
    do {
        reconnect_flag = 0;
        HTTPClient http;
      
        Serial.print("[HTTP] begin...\n");
        EnableDreqInt = true;        
        Decoding = false;
        
        http.begin(Url);
        http.addHeader("Accept-Encoding", "gzip, deflate, chunked, compress, identity");
        Serial.print("[HTTP] GET...\n");
        // start connection and send HTTP header
        int httpCode = http.GET();
        
        //if (httpCode > 0) {
        if (httpCode == HTTP_CODE_OK){
            // HTTP header has been send and Server response header has been handled
            //Serial.printf("[HTTP] GET... code: %d\n", httpCode);
            for (i=0; i<http.headers(); i++){
                Serial.printf("[HTTP] GET... %s=%s\n", http.headerName(i), http.header(i));
            }
            
            digitalWrite(POWER_LED, HIGH);
            
            // get length of document (is -1 when Server sends no Content-Length header)
            //int len = http.getSize();
            //Serial.printf("[HTTP] GET.. length: %d\n", len);
            
            // get tcp stream
            WiFiClient * stream = http.getStreamPtr();
            dat_start = 0;
            emp_start = 0;
            dat_size = 0;
            emp_size = RECVBUFF_SIZE - dat_size;
            
            // １バッファ分先読み
            get_sound_tm = millis();    // Timeoutの起点
            while(http.connected()) {
                // get available data size
                size_t size = stream->available();
                if (size > 0) {
                    Serial.printf("size=%d\n", size);
                    get_size = (size < emp_size ? size : emp_size);
                    data_read = stream->readBytes(RecvBuff + emp_start, get_size);
                    if (data_read > 0){
                        emp_start += data_read;
                        if (emp_start == RECVBUFF_SIZE){
                            emp_start = 0;
                        }
                        dat_size += data_read;
                        emp_size -= data_read;
                        //Serial.printf("dat_size=%d\n", dat_size);
                        if (emp_size == 0){
                            break;
                        }
                    }
                }
                if ((millis() - get_sound_tm) > 30*1000 ){
                    Serial.printf("timeout dat_size=%d\n", dat_size);
                    break;
                }
                delay(1);
            }
            if (emp_size > 0){        // timeout
                LOG_1("TIMEOUT");
                Serial.printf("TIMEOUT !!\n");
                http.end();
                // リトライしない
                //retry_cnt++;
                //if (retry_cnt < 3){
                //    reconnect_flag = 1;
                //}
                continue;
            }
#if 0           // sample rate, bit rate などの情報必要としない（以下のロジックはMP3専用）
            if (SampleRate == 0){
                for (i=0; i<(RECVBUFF_SIZE-2); i++){
                    if (RecvBuff[i] == 0xFF && (RecvBuff[i + 1] == 0xFB || RecvBuff[i + 1] == 0xFA)){
                        int bitrate_index = (RecvBuff[i + 2] >> 4);
                        int sample_index = ((RecvBuff[i + 2] & 0x0C) >> 2);
                        
                        if (sample_index >= 0 && sample_index <= 3){
                            FrameByte[0] = RecvBuff[i];
                            FrameByte[1] = RecvBuff[i + 1];
                            SampleRate = sample_tab[sample_index];
                            BitRate = bitrate_tab[bitrate_index];
                            SecFrame = (double)1152 / (double)SampleRate;
                            //save_frames = ((save_len_min * 60)*SampleRate) / 1152;
                            Serial.printf("SampleRate=%d Bitrate=%d sec=%f frames=%ld\n", SampleRate, BitRate, SecFrame, save_frames);
                            break;
                        }
                    }
                }
            }
#endif
            
            continue_flag = 1;
            get_sound_tm = millis();    // Timeoutの起点
            while (continue_flag == 1) {
                unsigned long   tm1 = millis();
                size_t size = stream->available();
                // TODO:available->readBytesの所要時間を計測すること
                
                if (size > 0 && dat_size < RECVBUFF_SIZE) {
                    emp_size = RECVBUFF_SIZE - dat_size;
                    if (emp_start + emp_size < RECVBUFF_SIZE){
                        get_size = emp_size;
                    } else {
                        get_size = RECVBUFF_SIZE - emp_start;
                    }
                    if (get_size > size) get_size = size;
                    data_read = stream->readBytes(RecvBuff + emp_start, get_size);
                    // ファイル出力
                    if (SaveAtDecode == true){
                        FileSnd.write(RecvBuff + emp_start, data_read);
                        //FileSnd.flush();
                        RecFileBytes += data_read;
                        //Serial.printf("write \n");
                    }
                    emp_start += data_read;
                    if (emp_start == RECVBUFF_SIZE){
                        emp_start = 0;
                    }
                    dat_size += data_read;
                    //get_sound_tm = millis();    // Timeoutの起点
                } else if (size == 0){
                    if (http.connected() == false){
                        LOG_1("disconnected");
                        Serial.printf("disconnected \n");
                        continue_flag = 0;
                        reconnect_flag = 1;
                    }
                }
                if (digitalRead(MP3_DREQ) != 0){
                    detachInterrupt(MP3_DREQ);
                    Decode();
                    attachInterrupt(MP3_DREQ, isr_dreq, RISING);
                }
                
                // SW
                if (EVENT_SWRETURN == true){
                    Serial.printf("SWRETURN\n");
                    EVENT_SWRETURN = false;
                    continue_flag = 0;
                } else if (EVENT_SWPOFF == true){
                    retval = SWPOFF_PRESSED;
                    EVENT_SWPOFF = false;
                    continue_flag = 0;
                } else if (EVENT_SWOPTION == true){
                    EVENT_SWOPTION = false;
                    if (SaveAtDecode == false && FileSnd){
                        detachInterrupt(MP3_DREQ);
                        if (dat_start < emp_start){
                            FileSnd.write(RecvBuff + dat_start, emp_start - dat_start);
                        } else {
                            FileSnd.write(RecvBuff + dat_start, RECVBUFF_SIZE - dat_start);
                            FileSnd.write(RecvBuff, emp_start);
                        }
                        attachInterrupt(MP3_DREQ, isr_dreq, RISING);
                        SaveAtDecode = true;
                        LCD_Println(1, LABEL_RECIEVE_RECORDING);
                    } else {
                        SaveAtDecode = false;
                        LCD_Println(1, LABEL_RECIEVE);
                    }
                }
                
            }
            
            Serial.println();
            Serial.print("[HTTP] connection closed or file end.\n");
        } else {
            Serial.printf("[HTTP] GET... failed, error: %s\n", http.errorToString(httpCode).c_str());
        }
        
        Serial.printf("http.end ...\n");
        http.end();
    } while (reconnect_flag == 1);
    
    LOG_1("recieve_streaming_listen END");
    EnableDreqInt = false;
    return retval;
}

/*
 *    ストリーミング方式　保存専用
 */
int recieve_streaming_schrec(char *Url, int save_len_min){
    int data_read;
    int get_size;
    int emp_size;
    unsigned long   start_tm;
    unsigned long   save_len_ms;
    int     i;
    int     continue_flag = 0;
    int     reconnect_flag;
    unsigned long   get_sound_tm;
    int        retry_cnt;
    char    buff[32];
    
    LOG_1("recieve_streaming_schrec START");
    LOG_1(Url);
    
    // VS1053使わない
    detachInterrupt(MP3_DREQ);
    digitalWrite(MP3_CS, HIGH);
    digitalWrite(MP3_DCS, HIGH);
    
    save_len_ms = save_len_min * 60 * 1000;       // 録音時間ミリ秒単位
    retry_cnt = 0;
    start_tm = millis();
    
    do {
        reconnect_flag = 0;
        HTTPClient http;
      
        Serial.print("[HTTP] begin...\n");
        http.begin(Url);
        http.addHeader("Accept-Encoding", "gzip, deflate, chunked, compress, identity");
        
        // start connection and send HTTP header
        int httpCode = http.GET();
        //if (httpCode > 0) {
        if (httpCode == HTTP_CODE_OK){
            for (i=0; i<http.headers(); i++){
                Serial.printf("[HTTP] GET... %s=%s\n", http.headerName(i), http.header(i));
            }
            
            // get tcp stream
            WiFiClient * stream = http.getStreamPtr();
            dat_start = 0;
            emp_start = 0;
            dat_size = 0;
            emp_size = RECVBUFF_SIZE - dat_size;
            
            continue_flag = 1;
            get_sound_tm = millis();    // Timeoutの起点
            retry_cnt = 0;
            
            do {
                size_t size = stream->available();
                if (size > 0 /*&& dat_size < RECVBUFF_SIZE*/) {
                    emp_size = RECVBUFF_SIZE - dat_size;
                    if (emp_start + emp_size < RECVBUFF_SIZE){
                        get_size = emp_size;
                    } else {
                        get_size = RECVBUFF_SIZE - emp_start;
                    }
                    if (get_size > size) get_size = size;
                    data_read = stream->readBytes(RecvBuff + emp_start, get_size);
                    dat_size += data_read;
                    emp_start += data_read;
                    if (dat_size == RECVBUFF_SIZE){
                        emp_start = 0;
                        dat_size = 0;
                        // ファイル出力
                        FileSnd.write(RecvBuff, RECVBUFF_SIZE);
                        FileSnd.flush();
                        RecFileBytes += RECVBUFF_SIZE;
                    }
                    get_sound_tm = millis();    // Timeoutの起点
                    retry_cnt = 0;
                } else if (size == 0){
                    if ((millis() - get_sound_tm) > GETSOUND_TIMEOUT){
                        sprintf(buff, "TIMEOUT (%ld)", RecFileBytes);
                        LOG_1(buff);
                        Serial.printf("timeout \n");
                        continue_flag = 0;
                        retry_cnt++;
                        if (retry_cnt < MAX_RETRY_COUNT){
                            reconnect_flag = 1;
                            wait_ms(retry_cnt * RETRY_WAIT_TIME);
                        }
                    } else if (http.connected() == false){
                        sprintf(buff, "disconnect (%ld)", RecFileBytes);
                        LOG_1(buff);
                        Serial.printf("disconnected \n");
                        continue_flag = 0;
                        retry_cnt++;
                        if (retry_cnt < MAX_RETRY_COUNT){
                            reconnect_flag = 1;
                            wait_ms(retry_cnt * RETRY_WAIT_TIME);   // リトライ多発のとき、少しずつ長い時間を置いてから再接続
                        }
                    }
                }
                
                // SW
                if (EVENT_SWRETURN == true){
                    Serial.printf("SWRETURN\n");
                    EVENT_SWRETURN = false;
                    continue_flag = 0;
                }
                // RTC timeout
                if ((millis() - start_tm) > save_len_ms){
                    continue_flag = 0;
                }
                
            } while (continue_flag == 1);
        } else {
            Serial.printf("[HTTP] GET... failed, error: %s\n", http.errorToString(httpCode).c_str());
        }
        Serial.printf("http.end ...\n");
        http.end();
    } while (reconnect_flag == 1);
    
    LOG_1("recieve_streaming_schrec END");
    return SWPOFF_PRESSED;
}

/*
 *    Bitrateの大きいサイト用
 */
int recieve_streaming_highbitrate(char *Url){
    int     reconnect_flag;
    int     retval;
    int     i, j;
    
    LOG_1("recieve_streaming_highbitrate START");
    retval = 0;
    detachInterrupt(MP3_DREQ);
    
    do {
        reconnect_flag = 0;
        HTTPClient http;
      
        Serial.print("[HTTP] begin...\n");
        http.begin(Url);
        int httpCode = http.GET();
        if (httpCode == HTTP_CODE_OK){
            //int len = http.getSize();
            //Serial.printf("[HTTP] GET.. length: %d\n", len);
            // get tcp stream
            WiFiClient * stream = http.getStreamPtr();
            do {
                size_t size = stream->available();
                if (size >= VS1053_DECODE_BYTE){
                    //Serial.printf("[HTTP] GET.. size: %d\n", size);        // 実測で最大4300ぐらい
                    //size &= 0xFFE0;
                    //int data_read = stream->readBytes(RecvBuff, size);    // 32の倍数バイトまとめてとる。
                    //j = 0;
                    Spi->beginTransaction(SPISettings(spiClk, MSBFIRST, SPI_MODE0));
                    do {
                        if (digitalRead(MP3_DREQ) != 0){
                            int data_read = stream->readBytes(RecvBuff, VS1053_DECODE_BYTE);    // 32バイトごととる
                            digitalWrite(MP3_DCS, LOW);
                            for (i=0; i<VS1053_DECODE_BYTE; i++){
                                //Spi->transfer(RecvBuff[j + i]);  
                                Spi->transfer(RecvBuff[i]);  
                            }
                            digitalWrite(MP3_DCS, HIGH);
                            //j += VS1053_DECODE_BYTE;
                            size -= VS1053_DECODE_BYTE;
                            // ファイル出力
                            if (SaveAtDecode == true){
                                FileSnd.write(RecvBuff, VS1053_DECODE_BYTE);
                                RecFileBytes += VS1053_DECODE_BYTE;
                            }
                        //} else {
                        //    Serial.printf("DREQ OFF\n");  // bitrate=320kではほとんど無い
                        }
                    } while (/*j < size*/ size >= VS1053_DECODE_BYTE);
                    Spi->endTransaction();
                } else if (size == 0){
                    if (http.connected() == false){
                        LOG_1("disconnected");
                        Serial.printf("disconnected \n");
                        reconnect_flag = 1;
                    }
                }
                
                // SW
                if (EVENT_SWUP == true){
                    EVENT_SWUP = false;
                    if (Volume <=(STEP_VOLUME + MIN_VOLUME)){
                        Volume = MIN_VOLUME;
                    } else {
                        Volume -= STEP_VOLUME;
                    }
                    VS1053_SetVolume(Volume);
                } else if (EVENT_SWDOWN == true){
                    EVENT_SWDOWN = false;
                    Volume += STEP_VOLUME;
                    if (Volume > MAX_VOLUME){
                        Volume = MAX_VOLUME;
                    }
                    VS1053_SetVolume(Volume);
                } else if (EVENT_SWRETURN == true){
                    Serial.printf("SWRETURN\n");
                    EVENT_SWRETURN = false;
                    break;
                } else if (EVENT_SWPOFF == true){
                    retval = SWPOFF_PRESSED;
                    EVENT_SWPOFF = false;
                    break;
                } else if (EVENT_SWOPTION == true){
                    EVENT_SWOPTION = false;
                    if (RecvMode == RECVMODE_LISTEN){
                        if (SaveAtDecode == false && FileSnd){
                            SaveAtDecode = true;
                            LCD_Println(1, LABEL_RECIEVE_RECORDING);
                        } else {
                            SaveAtDecode = false;
                            LCD_Println(1, LABEL_RECIEVE);
                        }
                    }
                }
            } while (1);
        }
        
        Serial.printf("http.end ...\n");
        http.end();
    } while (reconnect_flag == 1);
    
    attachInterrupt(MP3_DREQ, isr_dreq, RISING);
    LOG_1("recieve_streaming_highbitrate END");
    return retval;
}


/*
 *  デコード（イベント＆ポーリング方式）
 */
void Decode(){
    int data_read;
    int put_size;
    
    if (dat_size >= 32){
        if ((dat_start + dat_size) < RECVBUFF_SIZE){
            put_size = dat_size;
        } else {
            put_size = RECVBUFF_SIZE - dat_start;
        }
        data_read = VS1053_Decode(RecvBuff + dat_start, put_size);
        //if (SaveAtDecode == true){
        //    FileSnd.write(RecvBuff + dat_start, data_read);       // 割り込みハンドラからは実行できない
        //    FileSnd.flush();
        //    RecFileBytes += data_read;
        //}
        dat_start += data_read;
        if (dat_start == RECVBUFF_SIZE) dat_start = 0;
        dat_size -= data_read;
    }
    
    if (EVENT_SWUP == true){
        EVENT_SWUP = false;
        if (Volume <=(STEP_VOLUME + MIN_VOLUME)){
            Volume = MIN_VOLUME;
        } else {
            Volume -= STEP_VOLUME;
        }
        VS1053_SetVolume(Volume);
    } else if (EVENT_SWDOWN == true){
        EVENT_SWDOWN = false;
        Volume += STEP_VOLUME;
        if (Volume > MAX_VOLUME){
            Volume = MAX_VOLUME;
        }
        VS1053_SetVolume(Volume);
    }
}


/*
 *  ボリューム値をファイル保存
 */
void SaveVolume(){
    Serial.printf("SaveVolume %x\n", Volume);
    if (VolumeBak != Volume){
        FileObj = SD.open(VOLUME_FILE, FILE_WRITE);
        if(FileObj){
            Serial.printf("Opened\n");
            FileObj.write((uint8_t *)&Volume, sizeof(uint16_t));
            FileObj.close();
            VolumeBak = Volume;
            Serial.printf("Saved\n");
        } else {
            Serial.printf("volume file open error\n");
        }
    } else {
        Serial.printf("Volume no change\n");
    }
}

/*
 *  ボリューム値をファイルから取得
 */
void LoadVolume(){
    FileObj = SD.open(VOLUME_FILE);
    if (FileObj){
        FileObj.read((uint8_t *)&Volume, sizeof(uint16_t));
        FileObj.close();
        Serial.printf("LoadVolume %x\n", Volume);
        if (!(Volume > MIN_VOLUME && Volume < MAX_VOLUME)){
            Volume = DEFAULT_VOLUME;
        }
    } else {
        Volume = DEFAULT_VOLUME;
        Serial.printf("volume file open error\n");
    }
    
    VolumeBak = Volume;
    Serial.printf("LoadVolume %x\n", Volume);
}


/*
 *  M3Uファイルから音源ファイルを取得
 *  例(NHK-FM,改行=0A)
#EXTM3U
#EXT-X-VERSION:3
#EXT-X-TARGETDURATION:10
#EXT-X-MEDIA-SEQUENCE:2733599
#EXTINF:10.02667,
20210218T111826/master48k/01366/master48k_01599.aac
 ... 全１０件のaacファイル
 *
 *  例(i-wave,改行=0D0A)
#EXTM3U
#EXT-X-VERSION:3
#EXT-X-TARGETDURATION:11
#EXT-X-MEDIA-SEQUENCE:0
#EXT-X-PROGRAM-DATE-TIME:2022-01-02T00:57:14.000+00:00
#EXTINF:10.031,
https://chi-live3-ais-relay1.leanstream.co/musicbird/JCB048.stream/gG7RN2mG5w5p-164108496-10031.aac
 ... 全３件のaacファイル
 *
 */
void ParseM3U(char *m3u, int filesize, char nl, int nllen){
    char    *c1, *c2, *c3, *c4;
    int     i;
    
    Serial.printf("[ParseM3U]\n");
    c1 = m3u;
    
    while (1){
        c2 = strchr(c1, 0x0d);
        c3 = strchr(c1, 0x0a);
        if (c2 != NULL){
            *c2 = '\0';
            if (*(c2 + 1) == 0x0a){     // 改行=0d0a
                c4 = c2 + 2;            // c4に次の行の先頭
            } else {    // 改行=0d
                c4 = c2 + 1;
            }
        } else if (c3 != NULL){     // 改行=0a
            *c3 = '\0';
            c4 = c3 + 1;
        } else {
            break;
        } 
        if (*c1 != '#'){
            for (i=0; i<M3U_TABLE_SIZE; i++){
                if (strcmp(c1, M3uTable[i].reqfile) == 0){
                    break;
                }
            }
            if (i == M3U_TABLE_SIZE){
                strcpy(M3uTable[M3uTableInsIndex].reqfile, c1);
                M3uTableInsIndex++;
                if (M3uTableInsIndex == M3U_TABLE_SIZE){
                    M3uTableInsIndex = 0;
                }
                NumM3uTable++;
                if (NumM3uTable == M3U_TABLE_SIZE){
                    break;
                }
            }
        }
        c1 = c4;                // c1 は次の行の先頭
        if (*c1 == 0) break;    // ファイルの終端
        
    }
}

/*
 *  局リストのURLから取得できるM3Uファイルには別URLが記載されている
 * 例：
#EXTM3U
#EXT-X-STREAM-INF:BANDWIDTH=56000,CODECS="mp4a.40.5"
https://chi-live3-ais-relay1.leanstream.co/musicbird/JCB048.stream/playlist.m3u8?listeningSessionID=61bfec7700b942b8_1116098_idY9aERM_NjkuNjUuNDYuMzY6ODAwMA!!_00000003O6T&downloadSessionID=0
 *
 */
void ParseM3U_musicbird(char *m3u, int filesize, char nl, int nllen, char *Url){
    char    *c1, *c2;
    int     i;
    
    c1 = m3u;
    
    while (1){
        c2 = strchr(c1, nl);
        if (c2 == NULL) break;
        *c2 = '\0';         // 行の終端を設定
        if (*c1 != '#'){
            char    *c3;
            // https:をhttp:に変更
            c3 = strchr(c1, ':');
            if (c3 != NULL){
                strcpy(Url, "http");
                strcpy(Url + 4, c3);
            } else {
                strcpy(Url, c1);
            }
            break;
        }
        // c1-c2が1行
        c1 = c2 + nllen;            // c1 は次の行の先頭
        if (*c1 == 0) break;    // ファイルの終端
        
    }
}

/*-----------------------------------------------------------------------*/
/* VS1053                                                                */
/*-----------------------------------------------------------------------*/
/*
 *  VS1053初期化
 */
void VS1053_HardwareReset(){
    uint16_t  mode, modetmp;
    digitalWrite(MP3_RESET, HIGH);
    delay(250);
    digitalWrite(MP3_RESET, LOW);
    digitalWrite(MP3_CS, HIGH);
    digitalWrite(MP3_DCS, HIGH);
    delay(1000);
    digitalWrite(MP3_RESET, HIGH);

    while (digitalRead(MP3_DREQ) == LOW) ;
    //mode = SM_SDINEW_B;
    mode = SM_SDINEW_B | SM_DIFF_B;
    //mode = SM_SDINEW_B | SM_TESTS_B;
    VS1053_WriteSci(SCI_MODE, mode);
    Serial.println("VS1053_HardwareReset SCI_MODE");

    while (digitalRead(MP3_DREQ) == LOW) ;
    modetmp = VS1053_ReadSci(SCI_MODE);
    if (modetmp != mode){
        Serial.println("VS1053_HardwareReset SCI_MODE is not match");
    }

    while (digitalRead(MP3_DREQ) == LOW) ;
    //mode = 0x8000 + (6000);     // = 0x8000 + (12000000/2000)    // VS1011
    //mode = (0x8000 | CLOCKF_FREQ);     // = 0x8000 + ((12000000 - 8000000)/4000)  // VS1063
    mode = (0xc000 | SC_ADD | CLOCKF_FREQ);     // = 0x8000 + ((12000000 - 8000000)/4000) // VS1063
    VS1053_WriteSci(SCI_CLOCKF, mode);          // CLOCKF
    Serial.println("VS1053_HardwareReset SCI_CLOCKF");

    while (digitalRead(MP3_DREQ) == LOW) ;
    modetmp = VS1053_ReadSci(SCI_CLOCKF);
    if (modetmp != mode){
        Serial.println("VS1053_HardwareReset SCI_CLOCKF is not match");
    }

//    VS1053_LoadPatch();
    
    Serial.println("VS1053_HardwareReset END");
}

/*
 *  SCI書き込み
 */
void VS1053_WriteSci(uint8_t addr, uint16_t val){
    Spi->beginTransaction(SPISettings(spiClk, MSBFIRST, SPI_MODE0));
    digitalWrite(MP3_CS, LOW);
    Spi->transfer(2);  
    Spi->transfer(addr);  
    Spi->transfer(val >> 8);  
    Spi->transfer(val & 0x00FF);  
    digitalWrite(MP3_CS, HIGH);
    Spi->endTransaction();
}

/*
 *  SCI取得
 */
uint16_t VS1053_ReadSci(uint8_t addr){
    uint8_t   rbuf[4];
    uint16_t    val = 0;

    Spi->beginTransaction(SPISettings(spiClk, MSBFIRST, SPI_MODE0));
    digitalWrite(MP3_CS, LOW);
    rbuf[0] = Spi->transfer(3);
    rbuf[1] = Spi->transfer(addr);
    rbuf[2] = Spi->transfer(0);
    rbuf[3] = Spi->transfer(0);
    digitalWrite(MP3_CS, HIGH);
    Spi->endTransaction();

    val = rbuf[2];
    val <<= 8;
    val |= rbuf[3];
    return val;
}


/*
 *  vs1053b-patches.plgのLoadUserCode()を転用。
 *  データ plugin[] は vs1053b-patches.hで定義。
 */
#if 0
void VS1053_LoadPatch(){
  int i = 0;

  while (i<sizeof(plugin)/sizeof(plugin[0])) {
    unsigned short addr, n, val;
    addr = plugin[i++];
    n = plugin[i++];
    if (n & 0x8000U) { /* RLE run, replicate n samples */
      n &= 0x7FFF;
      val = plugin[i++];
      while (n--) {
        //WriteVS10xxRegister(addr, val);
        VS1053_WriteSci(addr, val);
      }
    } else {           /* Copy run, copy n samples */
      while (n--) {
        val = plugin[i++];
        //WriteVS10xxRegister(addr, val);
        VS1053_WriteSci(addr, val);
      }
    }
  }
}
#endif


/*
 *  VS1053に音源データを送付。送付できたバイト数を返す。
 */
int VS1053_Decode(uint8_t *buf, int size){
    int     i;
    int     remain = size;
    uint8_t *p = buf;
    
    Spi->beginTransaction(SPISettings(spiClk, MSBFIRST, SPI_MODE0));
    while (remain >= VS1053_DECODE_BYTE && digitalRead(MP3_DREQ) != 0){
        digitalWrite(MP3_DCS, LOW);
        for (i=0; i<VS1053_DECODE_BYTE; i++){
            Spi->transfer(*p++);  
        }
        digitalWrite(MP3_DCS, HIGH);
        remain -= VS1053_DECODE_BYTE;
    }
    Spi->endTransaction();
    
    return (size - remain);
}

void VS1053_SetVolume(uint16_t volume){
    uint16_t    vol;

    while(!digitalRead(MP3_DREQ));
    vol = (volume << 8) | volume;
    VS1053_WriteSci(SCI_VOL, vol);          // VOL
    
}

void VS1053_SineTest(){
    uint8_t   buff[16];
    int       i;

    delay(5000);    // テストモード実行のためにはVS1063の起動後に待ち時間がすこし必要
    buff[0] = 0x53;
    buff[1] = 0xEF;
    buff[2] = 0x6E;
    buff[3] = 126;      // 5168kHz
    buff[4] = 0x00;     // 最後に送る0x00の数はVS1011/VS1053とVS1063とで異なる
    buff[5] = 0x00;
    buff[6] = 0x00;
    buff[7] = 0x00;

    digitalWrite(MP3_DCS, LOW);
    Spi->beginTransaction(SPISettings(spiClk, MSBFIRST, SPI_MODE0));
    for (i=0; i<8; i++){
        Spi->transfer(buff[i]);  
    }
    Spi->endTransaction();
    digitalWrite(MP3_DCS, HIGH);
}

/*-----------------------------------------------------------------------*/
/* RTC M41T62                                                            */
/*-----------------------------------------------------------------------*/
uint8_t RTC_Init(){
    RTC_Read(0x0A, 1, RtcBuff);

    RtcBuff[0] &= 0xBF;         // SQWE off

    RTC_Write(0x0A, 1, RtcBuff);
    return 0;
}

void RTC_Reset(){
    uint8_t     ret;
    
    RtcBuff[0] = 0x80;      // 01h -> ST(stop bit) on
    RTC_Write(0x01, 1, RtcBuff);
    
    delay(10);
    
    RtcBuff[0] = 0x00;      // 01h -> ST off
    RTC_Write(0x01, 1, RtcBuff);
}

/*
 *    NTPサーバから時刻を取得してRTCにセット
 */
void RTC_SetNtpTime(){
    configTime(gmtOffset_sec, daylightOffset_sec, NtpServer /*"ntp1.jst.mfeed.ad.jp"*/ /*ntpServer*/);    // esp32-hal-time.c
    struct tm timeinfo;
    if (!getLocalTime(&timeinfo, 10000)){
        Serial.println("Failed to obtain time");
//        WiFi.disconnect();
        return;
    } else {
        Serial.printf("getLocalTime %d/%d/%d %d:%d \n", timeinfo.tm_year + 1900, timeinfo.tm_mon + 1, timeinfo.tm_mday, 
            timeinfo.tm_hour, timeinfo.tm_min);

        RtcBuff[0] = 0x00;      // 00h -> 0.1sec 0.01sec
        RtcBuff[1] = Dec2Bcd(timeinfo.tm_sec);           // 01h -> second
        RtcBuff[2] = Dec2Bcd(timeinfo.tm_min);           // 02h -> minute
        RtcBuff[3] = Dec2Bcd(timeinfo.tm_hour);          // 03h -> hour
        RtcBuff[4] = Dec2Bcd(timeinfo.tm_wday + 1);      // 04h -> day of week
        RtcBuff[5] = Dec2Bcd(timeinfo.tm_mday);          // 05h -> day of month
        RtcBuff[6] = Dec2Bcd(timeinfo.tm_mon + 1);       // 06h -> month
        RtcBuff[7] = Dec2Bcd(timeinfo.tm_year - 100);    // 07h -> year
        
        RTC_Write(0x00, 8, RtcBuff);
    }
//    WiFi.disconnect(true);
//    WiFi.mode(WIFI_OFF);
    
    
}

/*
 *  現在日時取得
 */
uint8_t RTC_GetCurrentDateTime(uint8_t *dt){
    uint8_t     ret;
    int         i;
    
    ret = RTC_Read(0x00, 8, RtcBuff);
    
    dt[5] = Bcd2Dec(RtcBuff[1] & 0x7F);    // sec
    dt[4] = Bcd2Dec(RtcBuff[2] & 0x7F);    // min
    dt[3] = Bcd2Dec(RtcBuff[3] & 0x3F);    // hour
    dt[2] = Bcd2Dec(RtcBuff[5] & 0x3F);    // day
    dt[1] = Bcd2Dec(RtcBuff[6] & 0x1F);    // month
    dt[0] = Bcd2Dec(RtcBuff[7]);             // year
    
    return ret;
}


/*
 *  RTC 書き込み
 */
uint8_t RTC_Write(uint8_t startaddress, uint8_t numbytes, uint8_t bytes[]){
    uint8_t     ret;
    int         i;
    
    Wire.beginTransmission(RTC_ADDRESS);
    Wire.write(startaddress);
    for (i=0; i<numbytes; i++){
        Wire.write(bytes[i]);
    }
    ret = Wire.endTransmission();
    //Serial.printf("endTransmission=%d\n", ret);
    return ret;
}

/*
 *  RTC 読み出し
 */
uint8_t RTC_Read(uint8_t startaddress, uint8_t numbytes, uint8_t bytes[]){
    int     i;
    
    RTC_Write(startaddress, 0, RtcBuff);
    Wire.requestFrom(RTC_ADDRESS, numbytes);
    for (i=0; i<numbytes; i++){
        bytes[i] = Wire.read();
    }
    
    return 0;
}

/*
 *  アラーム設定
 */
uint8_t RTC_SetAlarm(int mon, int date, int hour, int min, int sec){
    while (RTC_Read(0x0A, 6, RtcBuff) != 0) ;

    RtcBuff[0] &= 0x80;         // 0x0A : month (SQWE off)
    RtcBuff[1] &= 0xC0;         // 0x0B : date
    RtcBuff[2] &= 0x80;         // 0x0C : hour
    RtcBuff[3] &= 0x80;         // 0x0D : min
    RtcBuff[4] &= 0x80;         // 0x0E : sec
    RtcBuff[5] &= 0xFB;         // 0x0F : OF->OFF
    
    RtcBuff[0] |= Dec2Bcd(mon);
    RtcBuff[1] |= Dec2Bcd(date);
    RtcBuff[2] |= Dec2Bcd(hour);
    RtcBuff[3] |= Dec2Bcd(min);
    RtcBuff[4] |= Dec2Bcd(sec);
    
    // AFE/RPT bit
    RtcBuff[0] |= 0x80;     // AFE:1
    RtcBuff[1] &= 0x3F;     // RPT4,PRT5 : 0
    RtcBuff[2] &= 0x3F;     // RPT3 : 0
    RtcBuff[3] &= 0x7F;     // RPT2 : 0
    RtcBuff[4] &= 0x7F;     // RPT1 : 0

    while (RTC_Write(0x0A, 6, RtcBuff) != 0) ;
    
    return 0;
}

/*
 *  １０進数→BCD　変換
 */
uint8_t Dec2Bcd(uint8_t d){
    return ((d / 10) << 4) | (d % 10);
}

/*
 *  BCD→１０進数　変換
 */
uint8_t Bcd2Dec(uint8_t d){
    return ((d & 0xF0) >> 4) * 10 + (d & 0x0F);
}

/*
 *  AFビットの値(1/0)取得
 */
uint8_t RTC_GetAF(){
    while (RTC_Read(0x0F, 2, RtcBuff) != 0) ;
    if ((RtcBuff[0] & 0x40) != 0){
        return 1;
    } else {
        return 0;
    }
}

uint8_t RTC_ResetAlarm(){
    //while (RTC_Read(0x0A, 6, RtcBuff) != 0) ;
    while (RTC_Read(0x00, 1, RtcBuff) != 0) ;
    Serial.printf("Reset Alarm");
    return 0;
}

uint8_t RTC_ClearAlarm(){
    while (RTC_Read(0x0A, 6, RtcBuff) != 0) ;
    
    RtcBuff[0] = 0x00;      // 0x0A AFE:0
    RtcBuff[1] = 0x00;      // 0x0B
    RtcBuff[2] = 0x00;      // 0x0C
    RtcBuff[3] = 0x00;      // 0x0D
    RtcBuff[4] = 0x00;      // 0x0E
    RtcBuff[5] = 0x00;      // 0x0F
    
    while (RTC_Write(0x0A, 6, RtcBuff) != 0) ;
    
    return 0;
}

/*-----------------------------------------------------------------------*/
/* LCD                                                                   */
/*-----------------------------------------------------------------------*/

// LCDコマンド発行
void LCD_Cmd(byte x) {
  Wire.beginTransmission(LCD_ADDRESS);
  Wire.write(0b00000000); // CO = 0,RS = 0
  Wire.write(x);
  Wire.endTransmission();
}

// LCDセットアップ
void LCD_Setup(){
  LCD_Cmd(0b00111000); // function set
  LCD_Cmd(0b00111001); // function set
  LCD_Cmd(0b00000100); // EntryModeSet
  LCD_Cmd(0b00010100); // interval osc
  LCD_Cmd(0b01110000 | (LCD_CONTRAST & 0xF)); // contrast Low
  LCD_Cmd(0b01011100 | ((LCD_CONTRAST >> 4) & 0x3)); // contast High/icon/power
  LCD_Cmd(0b01101100); // follower control
  
  delay(200);
  
  LCD_Cmd(0b00111000); // function set
  LCD_Cmd(0b00001100); // Display On
  LCD_Cmd(0b00000001); // Clear Display
  delay(2);  
  
  LCD_SetContrast(30);
  delay(200);

}

void lcd_contdata(byte x) {
  Wire.write(0b11000000); // CO = 1, RS = 1
  Wire.write(x);
}
 
void lcd_lastdata(byte x) {
  Wire.write(0b01000000); // CO = 0, RS = 1
  Wire.write(x);
}
 
// 文字の表示
void LCD_PrintStr(const char *s) {
  Wire.beginTransmission(LCD_ADDRESS);
  while (*s) {
    if (*(s + 1)) {
      lcd_contdata(*s);
    } else {
      lcd_lastdata(*s);
    }
    s++;
  }
  Wire.endTransmission();
}

// LCD表示位置
// y:行(0-1) x:カラム(0-7)
void LCD_SetCursor(uint8_t x, uint8_t y) {
  LCD_Cmd(0x80 | (y * 0x40 + x));
}

// 行に文字列表示
void LCD_Println(int y, const char *s){
    LCD_SetCursor(0, y);
    LCD_PrintStr(s);
}

// LCDコントラスト
void LCD_SetContrast(uint8_t c) {
  LCD_Cmd(0x39);
  LCD_Cmd(0b01110000 | (c & 0x0f)); // contrast Low
  LCD_Cmd(0b01011100 | ((c >> 4) & 0x03)); // contast High/icon/power
  LCD_Cmd(0x38);
}

void wait_ms(unsigned long tm){
    unsigned long   tm1;
    
    tm1 = millis();
    while ((millis() - tm1) < tm){
        delay(1);
    }
}


/*
 *  cycle_ms : 周期(ms)
 */
void led_blink_start(int cycle_ms){
    BlinkLed = true;
    BlinkCycle2 = (cycle_ms / TICKER_INTERVAL);
    BlinkCycle = BlinkCycle2 / 2;
}

void led_blink_stop(){
    BlinkLed = false;
    digitalWrite(POWER_LED, HIGH);
}
